"""All prompts used for fact-checking subtasks prompting."""

# updates in Dec are donimant of function-based or code-based prompts, 
# to get rid of parsing LLM results
# ------------------------------------------------------------------------
# Dec 2023: decompose and decontextualise, return a list
# ------------------------------------------------------------------------
DOC_TO_INDEPEDENT_SENTENCES_PROMPT = """
Your task is to perform sentence segmentation and de-contextualization. 
Let's define a function named process(input:str).
The return value should be a list of strings, where each string should be a decontextualized sentence.
For example, if a user call process("Mary is a five-year old girl. She likes playing piano. She doesn't like cookies.").
You should return a python list without any other words, 
["Mary is a five-year old girl.", "Mary likes playing piano.", "Mary doesn't like cookies."]
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!

process("{doc}")
"""

SENTENCES_TO_CLAIMS_PROMPT = """
Your task is to decompose the text into atomic claims.
Let's define a function named decompose(input:str).
The returned value should be a list of strings, where each string should be a context-independent claim, representing one fact.
For example, if a user call decompose("Mary is a five-year old girl, she likes playing piano and she doesn't like cookies.").
You should return a python list without any other words: 
["Mary is a five-year old girl.", "Mary likes playing piano.", "Mary doesn't like cookies."]
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!

decompose("{doc}")
"""

# just sentence splits without decontextualization
DOC_TO_SENTENCES_PROMPT = """
Your task is to perform sentence segmentation. 
Let's define a function named split(input:str).
The return value should be a list of strings, where each string should be a sentence.
For example, if a user call process("Mary is a five-year old girl. She likes playing piano. She doesn't like cookies.").
You should return a python list without any other words, 
["Mary is a five-year old girl.", "Mary likes playing piano.", "Mary doesn't like cookies."]
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!

split("{doc}")
"""

# ------------------------------------------------------------------------
# Dec 2023: identify checkworthy
# ------------------------------------------------------------------------
CHECKWORTHY_PROMPT = """
Your task is to identify whether texts are checkworthy in the context of fact-checking.
Let's define a function named checkworthy(input: List[str]).
The return value should be a list of strings, where each string selects from ["Yes", "No"].
"Yes" means the text is a factual checkworthy statement.
"No" means that the text is not checkworthy, it might be an opinion, a question, or others.
For example, if a user call checkworthy(["I think Apple is a good company.", "Friends is a great TV series.", "Are you sure Preslav is a professor in MBZUAI?", "The Stanford Prison Experiment was conducted in the basement of Encina Hall.", "As a language model, I can't provide these info."])
You should return a python list without any other words, 
["No", "Yes", "No", "Yes", "No"]
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!


checkworthy({texts})
"""

CHECKWORTHY_PROMPT_BOOL = """
Your task is to identify whether texts are checkworthy in the context of fact-checking.
Let's define a function named checkworthy(input: List[str]).
The return value should be a list of bool values: [True, False].
True means the text is a factual checkworthy statement.
False means that the text is not checkworthy, it might be an opinion, a question, or others.
For example, if a user call checkworthy(["I think Apple is a good company.", "Friends is a great TV series.", "Are you sure Preslav is a professor in MBZUAI?", "The Stanford Prison Experiment was conducted in the basement of Encina Hall.", "As a language model, I can't provide these info."])
You should return a python list without any other words, 
[False, True, False, True, False]
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!


checkworthy({claims})
"""

SPECIFY_CHECKWORTHY_CATEGORY_PROMPT = """
You are a factchecker assistant with task to identify a sentence, whether it is 1. a factual claim; 2. an opinion; 3. not a claim (like a question or a imperative sentence); 4. other categories.
Let's define a function named checkworthy(input: str).
The return value should be a python int without any other words, representing index label, where index selects from [1, 2, 3, 4].

For example, if a user call checkworthy("I think Apple is a good company.")
You should return 2
If a user call checkworthy("Friends is a great TV series.")
You should return 1
If a user call checkworthy("Are you sure Preslav is a professor in MBZUAI?")
You should return 3
If a user call checkworthy("As a language model, I can't provide these info.")
You should return 4
Note that your response will be passed to the python interpreter, SO NO OTHER WORDS!

checkworthy("{sentence}")
"""

# ------------------------------------------------------------------------
# Dec 2023: Verify
# ------------------------------------------------------------------------
IDENTIFY_STANCE_PROMPT = """You are given a claim and an evidence text, and you need to decide whether the evidence supports, refutes, or is irrelevant to the claim. Choose from the following three options.
A. The evidence supports the claim.
B. The evidence refutes the claim.
C. The evidence is irrelevant to the claim.

For example, you are give Claim: "Preslav is a professor.", Evidence: "Preslav Nakov is a Professor in MBZUAI NLP group, and also the department chair." You should return A
Pick the correct option A, B, C without other words.

Claim: {claim}
Evidence: {evidence}"""

IDENTIFY_STANCE_PROMPT_FUNC = """
Let's define a function named verify(claim:str, evidence:str) -> {-1,0,1}
You are given a claim and an evidence text as input, and you need to decide whether the evidence supports, refutes, or is irrelevant to the claim. Choose from the following three options as the return value.
1: The evidence supports the claim.
-1: The evidence refutes the claim.
0: The evidence is irrelevant to the claim.

For example, when the user call verify(claim="Preslav is a professor.", evidence="Preslav Nakov is a Professor in MBZUAI NLP group, and also the department chair.")
You should return 1
Pick the correct option -1, 0, 1 without other words.

verify(claim="{claim}",evidence="{evidence}")"""


# , which correspond to the reasoning, whether the given text is factual or not (Boolean - True or False), the factual error present in the text, and the corrected text.

VERIFY_PROMPT = """
You are given a piece of text. Your task is to identify whether there are any factual errors within the text.
When you are judging the factuality of the given text, you could reference the provided evidences if needed. The provided evidences may be helpful. Some evidences may contradict to each other. You must be careful when using the evidences to judge the factuality of the given text.
The response should be a Python dict with four keys - "reasoning", "error", "correction" and "factuality".
The following is the given text:
[text]: {claim}
The following is the provided evidences:
[evidences]: {evidence}
You should only respond in format as described below. DO NOT RETURN ANYTHING ELSE. START YOUR RESPONSE WITH '{{'.
[response format]:
{{
    "reasoning": "Why is the given text factual or non-factual? Be careful when you said something is non-factual. When you said something is non-factual, you must provide multiple evidences to support your decision.",
    "error": "None if the text is factual; otherwise, describe the error in string.",
    "correction": "A string, the corrected text if there is an error.",
    "factuality": "An int value, 1 stands for the given text is factual, -1 is for non-factual, and 0 for irrelevant."
}}
"""
# ------------------------------------------
# Oct 2023
# ------------------------------------------
zero_shot_sentence_checkworthiness = """You are a factchecker assistant with task to identify sentences that are checkworthy. Sentence is checkworthy only if it contains factual claims.
Classify the check-worthiness of these sentences, and output the label yes or no:
{sentence}
output:
"""

zero_shot_claim_checkworthiness = """You are a factchecker assistant with task to identify a sentence, whether it is 1. a factual claim; 2. an opinion; 3. not a claim (like a question or a imperative sentence); 4. other categories. \n
Output the label index only: \n
{claim} \n
output:
"""

# We find that it is hard for model to distinguish complete support and partial support, merge as the one: support
zero_shot_claim_evidence_three_stance_prompt = "### Instruction: You are given a claim and an evidence text, and you need to decide whether the evidence supports, refutes, or is irrelevant to the claim.\n\n### Input:\n\nClaim: {claim}\n\nEvidence: {evidence}\n\nOptions are as follows:\n A) The evidence supports the claim.\n\n B) The evidence refutes the claim.\n C) The evidence is irrelevant to the claim.\n\n Pick the correct option. \n\n### Final Answer: "

zero_shot_claim_evidence_stance = """Given the evidence \n {evidence}, determine if the following statement is completely supported, partially supported, refuted or is irrelevant: {claim}, choose from four labels: 1. completely support, 2. partially support, 3. refute and 4. irrelevant.
Return the label index only.
Label index: 
"""

zero_shot_nli = """Given the premise sentence {}, determine if the following statement is entailed or contradicted or neutral: {}, by three labels: entailment, contradiction, neutral.
Label: 
"""

zero_shot_edit_response = """Given a document containing factual errors, please correct the errors in the document depending on a corresponding list of factually true claims. Note that preserve the linguistic features and style of the original document, just correct factual errors.

document: {response}

true claims: {claims}

revised document: """

zero_shot_edit_response_given_question = """Given a question, and an answer containing factual errors, please correct the errors in the document depending on a corresponding list of factually true claims. Note that preserve the linguistic features and style of the original document, just correct factual errors.

question: {prompt}

document: {response}

true claims: {claims}

revised document: """

# -------------------------------------------------------------------
# July 2023: decompose and decontextualise into atomic claims
# -------------------------------------------------------------------
# ZERO_SHOT_SENTENCE_TO_ATOMIC_CLAIMS = """Depending the context: {}, please breakdown the following sentence into independent facts and replace pronouns such as it, they, those, these, this, that, with specific entities or events.
# The sentence is: {}
# Atomic facts for this sentence are: """

ZERO_SHOT_SENTENCE_TO_ATOMIC_CLAIMS = """Depending the context: {}, please breakdown the following sentence into independent facts.
The sentence is: {}
Atomic facts for this sentence are: """

FEW_SHOT_SENTENCE_TO_ATOMIC_CLAIMS = """Depending the context, please breakdown the following sentence into independent facts.

Context: The United States has had two black presidents: Barack Obama, who served two terms from 2009 to 2017, and Donald Trump, who served one term from 2017 to 2021. Obama was the first black president in the history of the United States. He was born in Honolulu, Hawaii, to a mother from Kansas and a father from Kenya. Trump was the second black president. He was born in New York City and previously served as a businessman and reality television personality. 

The sentence is: The United States has had two black presidents: Barack Obama, who served two terms from 2009 to 2017, and Donald Trump, who served one term from 2017 to 2021.
Atomic facts for this sentence are: 
[
    "The United States has had two black presidents: Barack Obama and Donald Trump.",
    "Black president Barack Obama served two terms from 2009 to 2017.",
    "Black president Donald Trump served one term from 2017 to 2021."
]

The sentence is: Obama was the first black president in the history of the United States.
Atomic facts for this sentence are:
[
    "Obama was the first black president in the history of the United States."
]

The sentence is: He was born in Honolulu, Hawaii, to a mother from Kansas and a father from Kenya.
Atomic facts for this sentence are: 
[
    "Barack Obama was born in Honolulu, Hawaii.",
    "Barack Obama mother was from Kansas.",
    "Barack Obama father was from Kenya."
]

The sentence is: Trump was the second black president.
Atomic facts for this sentence are: 
[
    "Trump was the second black president."
]

The sentence is: He was born in New York City and previously served as a businessman and reality television personality.
Atomic facts for this sentence are: 
[
    "Donald Trump was born in New York City.",
    "Donald Trump previously served as a businessman",
    "Donald Trump previously served as a reality television personality."
]


Context: In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas. He was born on October 16, 1898, and served on the Supreme Court from 1939 until his retirement in 1975. Therefore, in 1980, Justice Douglas was still alive and would have been the oldest serving justice on the Court at that time.
The sentence is: In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas.
Atomic facts for this sentence are:
[
    "In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas."
] 

The sentence is: He was born on October 16, 1898, and served on the Supreme Court from 1939 until his retirement in 1975.
Atomic facts for this sentence are:
[
    "Justice William O. Douglas was born on October 16, 1898."
    "Justice William O. Douglas served on the Supreme Court from 1939 until his retirement in 1975."
] 

The sentence is: Therefore, in 1980, Justice Douglas was still alive and would have been the oldest serving justice on the Court at that time.
Atomic facts for this sentence are: 
[
    "Therefore, in 1980, Justice Douglas was still alive."
    "Justice William O. Douglas would have been the oldest serving justice on the Court in 1980."
]


Context: There have been only four female presidents of the United States in the country's history, so it is difficult to determine an average height for this group. The four female presidents were: \r\n1.Abigail Adams (1797-1801) \r\n2.Marilyn Carlson Nelson (2009-2013) \r\n3.Luci Baines Johnson (1973-1977) \r\n4.Hillary Clinton (2017-2021)
The sentence is: There have been only four female presidents of the United States in the country's history, so it is difficult to determine an average height for this group.
Atomic facts for this sentence are:
[
    "There have been only four female presidents of the United States in the country's history.",
    "It is difficult to determine an average height for four female presidents of the United States."
]

The sentence is: The four female presidents were: \r\n1.Abigail Adams (1797-1801) \r\n2.Marilyn Carlson Nelson (2009-2013) \r\n3.Luci Baines Johnson (1973-1977) \r\n4.Hillary Clinton (2017-2021)
Atomic facts for this sentence are:
[
    "Abigail Adams (1797-1801) is a female president of the United States.",
    "Marilyn Carlson Nelson (2009-2013) is a female president of the United States.",
    "Luci Baines Johnson (1973-1977) is a female president of the United States.",
    "Hillary Clinton (2017-2021) is a female president of the United States."
]


Context: {}
The sentence is: {}
Atomic facts for this sentence are:  
"""

# This prompt aims to break the document into decontextualised sentences, and then atomic claims
# Though it can not decontexlualize sentences, it can better break all sentences than the prompt above
# combined with using system_role = "You are good at document decomposition and decontextualization."
# date: 22/10/2023
FEW_SHOT_DECONTEXTUALIZE_SENTENCE_ATOMIC_CLAIMS = """Depending the context, please break it down into independent sentences, and breakdown the sentence into independent facts.
Context: The United States has had two black presidents: Barack Obama, who served two terms from 2009 to 2017, and Donald Trump, who served one term from 2017 to 2021. Obama was the first black president in the history of the United States. He was born in Honolulu, Hawaii, to a mother from Kansas and a father from Kenya. Trump was the second black president. He was born in New York City and previously served as a businessman and reality television personality. 

The sentence is: The United States has had two black presidents: Barack Obama, who served two terms from 2009 to 2017, and Donald Trump, who served one term from 2017 to 2021.
Atomic facts for this sentence are: 
[
    "The United States has had two black presidents: Barack Obama and Donald Trump.",
    "Black president Barack Obama served two terms from 2009 to 2017.",
    "Black president Donald Trump served one term from 2017 to 2021."
]

The sentence is: Obama was the first black president in the history of the United States.
Atomic facts for this sentence are:
[
    "Obama was the first black president in the history of the United States."
]

The sentence is: Barack Obama was born in Honolulu, Hawaii, to a mother from Kansas and a father from Kenya.
Atomic facts for this sentence are: 
[
    "Barack Obama was born in Honolulu, Hawaii.",
    "Barack Obama mother was from Kansas.",
    "Barack Obama father was from Kenya."
]

The sentence is: Trump was the second black president.
Atomic facts for this sentence are: 
[
    "Trump was the second black president."
]

The sentence is: Donald Trump was born in New York City and previously served as a businessman and reality television personality.
Atomic facts for this sentence are: 
[
    "Donald Trump was born in New York City.",
    "Donald Trump previously served as a businessman",
    "Donald Trump previously served as a reality television personality."
]


Context: In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas. He was born on October 16, 1898, and served on the Supreme Court from 1939 until his retirement in 1975. Therefore, in 1980, Justice Douglas was still alive and would have been the oldest serving justice on the Court at that time.
The sentence is: In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas.
Atomic facts for this sentence are:
[
    "In 1980, the oldest justice on the United States Supreme Court was Justice William O. Douglas."
] 

The sentence is: Justice William O. Douglas was born on October 16, 1898, and served on the Supreme Court from 1939 until his retirement in 1975.
Atomic facts for this sentence are:
[
    "Justice William O. Douglas was born on October 16, 1898."
    "Justice William O. Douglas served on the Supreme Court from 1939 until his retirement in 1975."
] 

The sentence is: Therefore, in 1980, Justice Douglas was still alive and would have been the oldest serving justice on the Court at that time.
Atomic facts for this sentence are: 
[
    "Therefore, in 1980, Justice Douglas was still alive."
    "Justice William O. Douglas would have been the oldest serving justice on the Court in 1980."
]


Context: There have been only four female presidents of the United States in the country's history, so it is difficult to determine an average height for this group. The four female presidents were: \r\n1.Abigail Adams (1797-1801) \r\n2.Marilyn Carlson Nelson (2009-2013) \r\n3.Luci Baines Johnson (1973-1977) \r\n4.Hillary Clinton (2017-2021)
The sentence is: There have been only four female presidents of the United States in the country's history, so it is difficult to determine an average height for this group.
Atomic facts for this sentence are:
[
    "There have been only four female presidents of the United States in the country's history.",
    "It is difficult to determine an average height for four female presidents of the United States."
]

The sentence is: The four female presidents were: \r\n1.Abigail Adams (1797-1801) \r\n2.Marilyn Carlson Nelson (2009-2013) \r\n3.Luci Baines Johnson (1973-1977) \r\n4.Hillary Clinton (2017-2021)
Atomic facts for this sentence are:
[
    "Abigail Adams (1797-1801) is a female president of the United States.",
    "Marilyn Carlson Nelson (2009-2013) is a female president of the United States.",
    "Luci Baines Johnson (1973-1977) is a female president of the United States.",
    "Hillary Clinton (2017-2021) is a female president of the United States."
]


Context: {}
The sentence is: {}
Atomic facts for this sentence are:  
"""

# -------------------------------------------------------------------
# April 2023: overall simple pipeline prompts
# -------------------------------------------------------------------
DECONTEXTILISATION_PROMPT = """Decompose and decontextualise a document into independently meaningful sentences. This process will make each sentence stand alone that can be verified independently.

Input: Mary is a five-year old girl. She likes playing piano. She doesn't like cookies. 
Output: 
Mary is a five-year old girl.
Mary likes playing piano.
Mary doesn't like cookies. 

Input: Google began as an online search firm, but it now offers more than 50 Internet services and products, from e-mail and online document creation to software for mobile phones and tablet computers. In addition, its 2012 acquisition of Motorola Mobility put it in the position to sell hardware in the form of mobile phones. 
Ouput: 
Google began as an online search firm.
Google now offers more than 50 Internet services and products.
Google offers from e-mail and online document creation to software for mobile phones and tablet computers.
Google 2012 acquisition of Motorola Mobility put it in the position to sell hardware in the form of mobile phones.

Input: """

CHECK_WORTHINESS_LABEL_ONLY_PROMPT = """Identify whether this claim is an opinion or factual, and whether it is checkworthy or not in the context of fact-checking. Just return two labels without explanation.
I think Apple is a good company.
opinon, not checkworthy
Preslav is a professor in MBZUAI.
factual, checkworthy
Friends is a great TV series.
opinion, not checkworthy
The Stanford Prison Experiment was conducted in the basement of Encina Hall.
factual, checkworthy
"""

ENTITY_EXTRACTION_PROMPT = """Extract all entities of a claim.
Input: Google now offers more than 50 Internet services and products.
Output: Google, Internet services, product
Input: Donald John Trump is an American politician, media personality, and businessman.
Output: Donald John Trump, American politician, media personality, businessman
Input: """

QGEN_PROMPT_DEP = """Give a list of queries using for searching related information for a claim. 
Input: Google now offers more than 50 Internet services and products. 
Output: What does Google offers now? 
How many service and product does Google offer? 
Google, more than 50 Internet services, products 
Input: Donald John Trump is an American politician, media personality, and businessman.
Output: Who is Donald John Trump?
Give information of Donald John Trump. 
Donald John Trump, American politician 
Donald John Trump, media personality 
Donald John Trump, businessman
Input: """

QGEN_PROMPT = """I will check things you said and ask questions.

You said: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
To verify it,
1. I googled: Does your nose switch between nostrils?
2. I googled: How often does your nostrils switch?
3. I googled: Why does your nostril switch?
4. I googled: What is nasal cycle?

You said: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
To verify it,
1. I googled: Where was Stanford Prison Experiment was conducted?

You said: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
To verify it,
1. I googled: What does Havel-Hakimi algorithm do?
2. I googled: Who are Havel-Hakimi algorithm named after?

You said: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.
To verify it,
1. I googled: Who sings the song "Time of My Life"?
2. I googled: Which film is the song "Time of My Life" from?
3. I googled: Who produced the song "Time of My Life"?

You said: Kelvin Hopins was suspended from the Labor Party due to his membership in the Conservative Party.
To verify it,
1. I googled: Why was Kelvin Hopins suspended from Labor Party?

You said: Social work is a profession that is based in the philosophical tradition of humanism. It is an intellectual discipline that has its roots in the 1800s.
To verify it,
1. I googled: What philosophical tradition is social work based on?
2. I googled: What year does social work have its root in?

You said: {claim}
To verify it,
""".strip()

QGEN_PROMPT_FMT = '''
You need to ask N questions based on the provided claim.
Here are some examples:
- Claim:
Social work is a profession that is based in the philosophical tradition of humanism. It is an intellectual discipline that has its roots in the 1800s.
- N=4
- Questions you may response:
["Does your nose switch between nostrils?", "How often does your nostrils switch?", "Why does your nostril switch?", "What is nasal cycle?"]

- Claim:
The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
- N=1
- Questions you may response:
["Where was Stanford Prison Experiment was conducted?"]

- Claim:
The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
- N=2
- Questions you may response:
["What does Havel-Hakimi algorithm do?", "Who are Havel-Hakimi algorithm named after?"]

Remember, you need to put your questions into a python list so that I will search them with the search engine API, so DON'T RETURN ANY OTHER IRRELEVANT WORDS!
- Claim:
{claim}
- N={n}
'''.strip()

STANCE_DETECTION_PROMPT = """Determine whether the evidence support the claim or not. Choose label from [support, partial support, refute, other] and explain why. 
Support means we can entail the claim by the evidence.
Partial support means: part of the information presented in the claim appear in the evidence.
Refute means that the evidence mention the same event as the claim, but a clear opposite fact. It should be highlighed that under refute, the evidence mentions the fact in the claim, they are closely relevant, but opposite meaning or stance.
Other means the evidence does not mention anything about the fact described in the claim, such that it neither supports nor refutes the claim.

Claim: Elon Musk is the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: support, statement 'he is also the founder, CEO and chief engineer of SpaceX' in evidence above supports the claim.

Claim: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the founder, CEO and chief engineer of SpaceX.
Stance: partial support.

Claim: Steve Jobs is the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: refute.

Claim: Elon Musk is a professor in The Stanford University.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: other, according to the evidence, I cannot judge whether the claim is true or not, not enough information, the evidence neither supports nor refutes.

Claim: On January 6, 2021, a mob of supporters of former President Donald Trump stormed the U.S. Capitol in an attempt to overturn the 2020 presidential election.
Evidence: On January 6, 2021, following the defeat of U.S. President Donald Trump in the 2020 presidential election, a mob of his supporters attacked the United States Capitol Building in Washington, D.C. The mob sought to keep Trump in power by preventing a joint session of Congress from counting the electoral college votes to formalize the victory of President-elect Joe Biden.
Stance: support.

Claim: The 2021 Capitol Hill riots resulted in the deaths of five people, including a Capitol police officer. 
Evidence: Five people died either shortly before, during, or following the riot: one was shot by Capitol Police, another died of a drug overdose, and three died of natural causes.
Stance: partial support, the evidence supports that fact that five deaths, but not sure whether they include a Capitol police officer or not.

Claim: More than 300 people have been charged with crimes related to the riots. 
Evidence: As of November 10, 2022, over 940 people had been charged in the Capitol breach. 
Stance: refute, evidence and claim are describing the same thing, the number of people who was charged is over 940, while more than 300 in the claim, so the evidence refutes the claim.

Claim: More than 300 people have been charged with crimes related to the riots. 
Evidence: The laptop computer taken from Pelosi's office was taken by 22-year-old Capitol rioter Riley Williams. Williams was arrested and indicted on eight counts, including theft of government property, obstructing an official proceeding, and assaulting or resisting police.
Stance: other, the evidence demonstrates something relevent to the fact in the claim, but it does not support or refute any information of it. 

Claim: {}
Evidence: {}
Stance: """

EDITOR_PROMPT = """Fix the claim according to the evidence.

Claim: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
Evidence: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
This suggests 45 minutes switch time in your statement is wrong.
Fix: Your nose switches back and forth between nostrils. When you sleep, you switch about every 2 hours. This is to prevent a buildup of mucus. It’s called the nasal cycle.

Claim: In the battles of Lexington and Concord, the British side was led by General Thomas Hall.
Evidence: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
This suggests General Thomas Hall in your statement is wrong.
Fix: In the battles of Lexington and Concord, the British side was led by Lieutenant Colonel Francis Smith.

Claim: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
Evidence: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
This suggests Encina Hall in your statement is wrong.
Fix: The Stanford Prison Experiment was conducted in the basement of Jordan Hall, Stanford’s psychology building.

Claim: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
Evidence: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
This suggests the Havel-Hakimi algorithm’s functionality in your statement is wrong.
Fix: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. It is named after Vaclav Havel and Samih Hakimi.

Claim: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Phil Ramone.
Evidence: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
This suggests "Time of My Life" producer name in your statement is wrong.
Fix: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.

Claim: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 1.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.
Evidence: Phoenix Market City was opened in January 2013 and has the distinction of being the largest mall in the city of Pune, with the area of 3.4 million square feet. It is located in the Viman Nagar area of Pune.
This suggests the 1.4 million square feet of built-up space in your statment is wrong.
Fix: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 3.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.

Claim: {claim}
Evidence: {evidence}
This suggests
""".strip()



from .rarr_utils import agreement_gate
from .prompts import rarr_prompts

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver


@Solver.register("rarr_agreement_gate", "claims_with_evidences", "claims_with_gates")
class RARRAgreementGate(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_evidences_per_question = args.get("max_evidences_per_question", 1)
        self.model = self.global_config.get("model", "gpt-4o-instruct")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        for claim, contents in claims.items():
            context = contents.get("context", None)
            evidences = contents.get("evidences", [])[: self.max_evidences_per_question]
            gates = []
            for evidence in evidences:
                gate = agreement_gate.run_agreement_gate(
                    claim=claim,
                    context=context,
                    query=evidence["query"],
                    evidence=evidence["text"],
                    model=self.model,
                    prompt=rarr_prompts.CONTEXTUAL_AGREEMENT_GATE_PROMPT
                    if context
                    else rarr_prompts.AGREEMENT_GATE_PROMPT,
                )
                gates.append(gate)
            contents["gates"] = gates

        state.set(self.output_name, claims)
        return True, state



import os
import uuid
import tqdm
import json
import traceback
from typing import Callable

from openfactcheck.utils.logging import get_logger
from openfactcheck.base import OpenFactCheck
from openfactcheck.state import FactCheckerState

# Get the logger
logger = get_logger(__name__)

class ResponseEvaluator:
    """
    This class is used to evaluate the factuality of a response using the pipeline of solvers.
    """
    def __init__(self, ofc: OpenFactCheck):
        """
        Initialize the ResponseEvaluator object.
        """
        
        # Set the OpenFactCheck object
        self.ofc = ofc

    def persist_output(self, state: FactCheckerState, idx, solver_name, cont, sample_name=0):
        """
        Persist the output of the solver
        """
        result = {
            "idx": idx,
            "solver": solver_name,
            "continue": cont,
            "state": state.to_dict()
        }

        # Create the output path
        output_path = os.path.join(self.ofc.output_path, os.path.dirname(sample_name))
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        # Write the output to a file
        with open(os.path.join(self.ofc.output_path, f'{sample_name}.jsonl'), 'a', encoding="utf-8") as f:
            f.write(json.dumps(result, ensure_ascii=False) + '\n')

    def read_output(self, sample_name):
        """
        Read the output file for the given sample
        """
        with open(os.path.join(self.ofc.output_path, f'{sample_name}.jsonl'), 'r', encoding="utf-8") as f:
            return [json.loads(line) for line in f]
        
    def remove_output(self, sample_name):
        """
        Remove the output file for the given sample
        """
        os.remove(os.path.join(self.ofc.output_path, f'{sample_name}.jsonl'))

    def evaluate(self, response: str, question: str = None, callback: Callable = None, **kwargs):
        """
        Evaluate the response using the pipeline and return the output
        """

        # Check if sample_name is provided in kwargs else generate a random one
        sample_name = kwargs.get("sample_name", str(uuid.uuid4()))

        # Initialize the state
        solver_output = FactCheckerState(question=question, response=response)

        # Initialize the output name
        output_name = "response"
        for idx, (name, (solver, input_name, output_name)) in tqdm.tqdm(enumerate(self.ofc.pipeline.items()),
                                                            total=len(self.ofc.pipeline)):
            logger.info(f"Invoking solver: {idx}-{name}")
            logger.debug(f"State content: {solver_output}")
        
            try:
                # Solver input is the output of the previous solver
                solver_input = solver_output

                # Run the solver
                cont, solver_output = solver(solver_input, **kwargs)

                # Persist the output
                logger.debug(f"Latest result: {solver_output}")
                if callback:
                    callback(
                        index=idx,
                        sample_name=sample_name,
                        solver_name=name,
                        input_name=input_name,
                        output_name=output_name,
                        input=solver_input.__dict__,
                        output=solver_output.__dict__,
                        continue_run=cont
                    )
                
                self.persist_output(solver_output, idx, name, cont, sample_name=sample_name)
                
            except:
                logger.error(f"Error at {traceback.format_exc()}")
                cont = False
                output_name = input_name

            # Break if the solver returns False
            if not cont:
                logger.info(f"Break at {name}")
                break

        return solver_output.get(output_name)

    def evaluate_streaming(self, response: str, question: str = None, **kwargs):
        """
        Evaluate the response using the pipeline and stream the output
        """

        def evaluate_response():
            # Check if sample_name is provided in kwargs else generate a random one
            sample_name = kwargs.get("sample_name", str(uuid.uuid4()))

            # Initialize the state
            solver_output = FactCheckerState(question=question, response=response)

            # Initialize the output name
            output_name = "response"
            for idx, (name, (solver, input_name, output_name)) in tqdm.tqdm(enumerate(self.ofc.pipeline.items()),
                                                                total=len(self.ofc.pipeline)):
                logger.info(f"Invoking solver: {idx}-{name}")
                logger.debug(f"State content: {solver_output}")
            
                try:
                    # Solver input is the output of the previous solver
                    solver_input = solver_output

                    # Run the solver
                    cont, solver_output = solver(solver_input, **kwargs)

                    # Persist the output
                    logger.debug(f"Latest result: {solver_output}")
                    
                    # Stream the output
                    yield {
                        "index": idx,
                        "solver_name": name,
                        "input_name": input_name,
                        "output_name": output_name,
                        "input": solver_input.__dict__,
                        "output": solver_output.__dict__,
                        "continue_run": cont
                    }

                    self.persist_output(solver_output, idx, name, cont, sample_name=sample_name)
                    
                except:
                    logger.error(f"Error at {traceback.format_exc()}")
                    cont = False
                    output_name = input_name

                # Break if the solver returns False
                if not cont:
                    logger.info(f"Break at {name}")
                    break
        
        # Execute the generator if stream is True, otherwise process normally
        return evaluate_response()


from transformers import AutoTokenizer, AutoModelForSequenceClassification, pipeline

# Load model directly
# Sentiment analysis pipeline
# classifier = pipeline("sentiment-analysis", model="roberta-large-mnli")

tokenizer = AutoTokenizer.from_pretrained("roberta-large-mnli")
model = AutoModelForSequenceClassification.from_pretrained("roberta-large-mnli")
classifier = pipeline('sentiment-analysis', model=model, tokenizer=tokenizer)

nli_labelmap = {
    "NEUTRAL": 3,
    "CONTRADICTION":2,
    "ENTAILMENT": 1   
}

nli2stance = {
    "NEUTRAL": 0,
    "CONTRADICTION": -1,
    "ENTAILMENT": 1   
}

stance_map = {
    'irrelevant': 3,
    'refute': 2,
    'partially-support': 1,
    'completely-support': 1
}

def nli_infer(premise, hypothesis):
    # predict one example by nli model
    try: 
        input = "<s>{}</s></s>{}</s></s>".format(premise, hypothesis)
        pred = classifier(input)
        # print(pred)
    except:
        # token length > 514
        L = len(premise)
        premise = premise[:int(L/2)]
        input = "<s>{}</s></s>{}</s></s>".format(premise, hypothesis)
        pred = classifier(input)
        # print(pred) 
        # [{'label': 'CONTRADICTION', 'score': 0.9992701411247253}]
    return nli2stance[pred[0]['label']]


from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .rarr_utils.question_generation import run_rarr_question_generation
from .rarr_utils.functional_prompt import QGEN_PROMPT
from .rarr_utils import search


@Solver.register("rarr_retriever", "claims", "claims_with_evidences")
class RARRRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("rarr_model", "gpt-4o-instruct")
        self.temperature_qgen = args.get("temperature_qgen", 0.7)
        self.num_rounds_qgen = args.get("num_rounds_qgen", 3)
        self.max_search_results_per_query = args.get("max_search_results_per_query", 5)
        self.max_sentences_per_passage = args.get("max_sentences_per_passage", 4)
        self.sliding_distance = args.get("sliding_distance", 1)
        self.max_passages_per_search_result = args.get("max_passages_per_search_result", 1)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        results = dict()
        for claim in claims:
            questions = run_rarr_question_generation(
                claim=claim,
                context=None,
                model=self.model,
                prompt=QGEN_PROMPT,
                temperature=self.temperature_qgen,
                num_rounds=self.num_rounds_qgen,
            )
            evidences = []
            for question in questions:
                q_evidences = search.run_search(
                    query=question,
                    max_search_results_per_query=self.max_search_results_per_query,
                    max_sentences_per_passage=self.max_sentences_per_passage,
                    sliding_distance=self.sliding_distance,
                    max_passages_per_search_result_to_return=self.max_passages_per_search_result,
                )
                evidences.extend([(question, x["text"]) for x in q_evidences])

            results[claim] = evidences

        state.set(self.output_name, results)
        return True, state



from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .factool_utils.chat_api import OpenAIChat
from .factool_utils.search_api import GoogleSerperAPIWrapper
from .factool_utils.prompt import QUERY_GENERATION_PROMPT


@Solver.register("factool_retriever", "claims", "claims_with_evidences")
class FactoolRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.gpt_model = self.global_config.get("factool_gpt_model", "gpt-4o")
        self.snippet_cnt = args.get("snippet_cnt", 10)
        self.gpt = OpenAIChat(self.gpt_model)
        self.query_prompt = QUERY_GENERATION_PROMPT
        self.search_engine = GoogleSerperAPIWrapper(snippet_cnt=self.snippet_cnt)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        queries = self._query_generation(claims=claims)
        evidences = self.search_engine.run(queries)
        results = {}
        for query, claim, evidence in zip(queries, claims, evidences):
            merged_query = " ".join(query) if query and len(query) > 1 else str(query) if query else ""
            results[claim] = [(merged_query, x["content"]) for x in evidence]
        state.set(self.output_name, results)
        return True, state

    def _query_generation(self, claims):
        messages_list = [
            [
                {"role": "system", "content": self.query_prompt["system"]},
                {
                    "role": "user",
                    "content": self.query_prompt["user"].format(input=claim),
                },
            ]
            for claim in claims
        ]
        return self.gpt.run(messages_list, list)



from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("factool_post_editor", "claim_info", "claim_info")
class FactoolPostEditor(StandardTaskSolver):
    """
    A solver to post-process the results of the Factool model. 
    Used to presents the results in human-readable format and to save the analysis in a JSON file.
    """
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        for key, pair in claim_info.items():
            claim_info[key]['edited_claims'] = claim_info[key]['stances'][0]

        state.set(self.output_name, claim_info)
        return True, state



from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("all_pass_abstain_detector", "response", "response")
class AllPassAbstainDetector(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        return True, state



from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .factool_utils.chat_api import OpenAIChat
from .factool_utils.prompt import VERIFICATION_PROMPT


@Solver.register("factool_verifier", "claims_with_evidences", "label")
class FactoolVerifier(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.gpt_model = self.global_config.get("factool_gpt_model", "gpt-4o")
        self.gpt = OpenAIChat(self.gpt_model)
        self.verification_prompt = VERIFICATION_PROMPT

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims_with_evidences = state.get(self.input_name)
        results = self._verification(claims_with_evidences)
        for i, k in enumerate(list(claims_with_evidences.keys())):
            results[i]["claim"] = k
            results[i]["evidences"] = claims_with_evidences[k]
        state.set("detail", results)
        label = all(v["factuality"] for v in results)
        state.set(self.output_name, label)
        return True, state

    def _verification(self, claims_with_evidences):
        messages_list = [
            [
                {"role": "system", "content": self.verification_prompt["system"]},
                {
                    "role": "user",
                    "content": self.verification_prompt["user"].format(
                        claim=claim, evidence=str([e[1] for e in evidence])
                    ),
                },
            ]
            for claim, evidence in claims_with_evidences.items()
        ]
        return self.gpt.run(messages_list, dict)



import bs4
import spacy
import requests
from collections import Counter
from string import punctuation
from typing import List, Dict, Tuple, Any


def is_tag_visible(element: bs4.element) -> bool:
    """Determines if an HTML element is visible.

    Args:
        element: A BeautifulSoup element to check the visiblity of.
    returns:
        Whether the element is visible.
    """
    if element.parent.name in [
        "style",
        "script",
        "head",
        "title",
        "meta",
        "[document]",
    ] or isinstance(element, bs4.element.Comment):
        return False
    return True


def scrape_url(url: str, timeout: float = 3) -> Tuple[str, str]:
    """Scrapes a URL for all text information.

    Args:
        url: URL of webpage to scrape.
        timeout: Timeout of the requests call.
    Returns:
        web_text: The visible text of the scraped URL.
        url: URL input.
    """
    # Scrape the URL
    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()
    except requests.exceptions.RequestException as _:
        print("URL Require Error.")
        return None, url

    # Extract out all text from the tags
    try:
        soup = bs4.BeautifulSoup(response.text, "html.parser")
        texts = soup.findAll(text=True)
        # Filter out invisible text from the page.
        visible_text = filter(is_tag_visible, texts)
    except Exception as _:
        print("BS4 Error.")
        return None, url

    # Returns all the text concatenated as a string.
    web_text = " ".join(t.strip() for t in visible_text).strip()
    # Clean up spacing.
    web_text = " ".join(web_text.split())
    return web_text, url


def get_hotwords(text: str, top_k: int = 10) -> List[str]:
    """# extract key words for a text, return most frequent topk keywords
    """
    nlp = spacy.load("en_core_web_sm")
    pos_tag = ['PROPN', 'ADJ', 'NOUN'] 
    doc = nlp(text.lower()) 

    result = []
    for token in doc:
        if(token.text in nlp.Defaults.stop_words or token.text in punctuation):
            continue
        if(token.pos_ in pos_tag):
            result.append(token.text)
            
    most_common_list = Counter(result).most_common(top_k)
    keywords = [item[0] for item in most_common_list]
    return keywords


def select_doc_by_keyword_coverage(claim: str, docs: List[str], 
                                   top_k_keywords: int = 10, top_k_docs: int = 5) -> List[int]:
    """count how many keywords appeared in this document len(appeared_keywords)
       sort documents by the count that represents the degree of coverage of the claim for the doc
       return index of top-k docs"""
    # get keywords in the claim.
    keywords = get_hotwords(claim, top_k_keywords)
    
    # how many keywords are contained in each doc
    counts = []
    for doc in docs:
        doc = doc.lower() # as all keywords are lowercase
        count = [1 for word in keywords if word in doc]
        counts.append(sum(count))

    # we keep the docs that contain the most keywords, as we aim to cut off lots of unrelevant docs
    max_count = max(counts)
    selected_docs_index = [i for i in range(len(docs)) if counts[i] == max_count]
    if len(selected_docs_index) < top_k_docs:
        # we sort docs by coverage, then keep top-K
        docs_index_sorted_coverage = sorted(range(len(counts)), key=lambda k: counts[k], reverse=True)
        selected_docs_index = docs_index_sorted_coverage[:top_k_docs]
    
    print("There are {} web pages selected.".format(len(selected_docs_index)))
    return selected_docs_index


def chunk_text(text: str, sentences_per_passage: int, 
               filter_sentence_len: int, sliding_distance: int = None) -> List[str]:
    """Chunks text into passages using a sliding window.

    Args:
        text: Text to chunk into passages.
        sentences_per_passage: Number of sentences for each passage.
        filter_sentence_len: Maximum number of chars of each sentence before being filtered.
        sliding_distance: Sliding distance over the text. Allows the passages to have
            overlap. The sliding distance cannot be greater than the window size.
    Returns:
        passages: Chunked passages from the text.
    """
    TOKENIZER = spacy.load("en_core_web_sm", disable=["ner", "tagger", "lemmatizer"])
    if not sliding_distance or sliding_distance > sentences_per_passage:
        sliding_distance = sentences_per_passage
    assert sentences_per_passage > 0 and sliding_distance > 0

    passages = []
    try:
        doc = TOKENIZER(text[:500000])  # Take 500k chars to not break tokenization.
        sents = [
            s.text
            for s in doc.sents
            if len(s.text) <= filter_sentence_len  # Long sents are usually metadata.
        ]
        for idx in range(0, len(sents), sliding_distance):
            passages.append(" ".join(sents[idx : idx + sentences_per_passage]))
    except UnicodeEncodeError as _:  # Sometimes run into Unicode error when tokenizing.
        print("Unicode error when using Spacy. Skipping text.")

    return passages


def select_passages_by_semantic_similarity(claim: str, selected_docs: List[str],
                                           max_sentences_per_passage: int = 3, filter_sentence_len: int = 250,
                                           sliding_distance: int = 3, top_k_passage: int = 5) -> Tuple[list, list]:
    passages: List[str] = []
    for doc in selected_docs:
        # RARR default setting (5, 250, 1) for chunk
        snippets = chunk_text(doc, max_sentences_per_passage, filter_sentence_len, sliding_distance) 
        passages.extend(snippets)
    passages = list(set(passages)) # remove repeated ones
    print("{} snippets of text are splitted.".format(len(passages)))
    
    # score each snippet of text against claim
    nlp = spacy.load("en_core_web_sm")
    claim = nlp(claim)
    sim = []
    for p in passages:
        sim.append(claim.similarity(nlp(p)))
    
    # sort by similarity score and keep topk
    index_sorted_sim = sorted(range(len(sim)), key=lambda k: sim[k], reverse=True)
    topk_passages = [passages[i] for i in index_sorted_sim[:top_k_passage]]

    # find docs of topk_passages: one passage may occur in multiple docs
    passage_doc_id: List[list] = []
    for p in topk_passages:
        temp = []
        for id, doc in enumerate(selected_docs):
            if p in doc:
                temp.append(id)
        # if fail to find docs of this passage, just pass.
        # this will lead some [], [], [] in evidence list for this snippet of text 
        if len(temp) == 0:
            print("Error in matching selected passage to its docs!")
        passage_doc_id.append(temp)

    return topk_passages, passage_doc_id


"""All prompts used for fact-checking subtasks prompting."""

CLAIM_EXTRACTION_PROMPT = {
    "system": "Please provide the claim you would like to fact-check.",
    "user": """You are given a piece of text that includes knowledge claims. A claim is a statement that asserts something as true or false, which can be verified by humans. Your task is to accurately identify and extract every claim stated in the provided text. Then, resolve any coreference (pronouns or other referring expressions) in the claim for clarity. Each claim should be concise (less than 15 words) and self-contained.
Your response MUST be a list of dictionaries. Each dictionary should contains the key "claim", which correspond to the extracted claim (with all coreferences resolved).
You MUST only respond in the format as described below. DO NOT RESPOND WITH ANYTHING ELSE. ADDING ANY OTHER EXTRA NOTES THAT VIOLATE THE RESPONSE FORMAT IS BANNED. START YOUR RESPONSE WITH '['.
[response format]: 
[
    {{
    "claim": "Ensure that the claim is fewer than 15 words and conveys a complete idea. Resolve any coreference (pronouns or other referring expressions) in the claim for clarity",
    }},
    ...
]

Here are two examples:
[text]: Tomas Berdych defeated Gael Monfis 6-1, 6-4 on Saturday. The sixth-seed reaches Monte Carlo Masters final for the first time . Berdych will face either Rafael Nadal or Novak Djokovic in the final.
[response]: [{{"claim": "Tomas Berdych defeated Gael Monfis 6-1, 6-4"}}, {{"claim": "Tomas Berdych defeated Gael Monfis 6-1, 6-4 on Saturday"}}, {{"claim": "Tomas Berdych reaches Monte Carlo Masters final"}}, {{"claim": "Tomas Berdych is the sixth-seed"}}, {{"claim": "Tomas Berdych reaches Monte Carlo Masters final for the first time"}}, {{"claim": "Berdych will face either Rafael Nadal or Novak Djokovic"}}, {{"claim": "Berdych will face either Rafael Nadal or Novak Djokovic in the final"}}]

[text]: Tinder only displays the last 34 photos - but users can easily see more. Firm also said it had improved its mutual friends feature.
[response]: [{{"claim": "Tinder only displays the last photos"}}, {{"claim": "Tinder only displays the last 34 photos"}}, {{"claim": "Tinder users can easily see more photos"}}, {{"claim": "Tinder said it had improved its feature"}}, {{"claim": "Tinder said it had improved its mutual friends feature"}}]

Now complete the following,ONLY RESPONSE IN A LIST FORMAT, NO OTHER WORDS!!!:
[text]: {input}
[response]: 
"""
}

QUERY_GENERATION_PROMPT = {
    "system": "You are a query generator that generates effective and concise search engine queries to verify a given claim. You only response in a python list format(NO OTHER WORDS!)",
    "user": """You are a query generator designed to help users verify a given claim using search engines. Your primary task is to generate a Python list of two effective and skeptical search engine queries. These queries should assist users in critically evaluating the factuality of a provided claim using search engines.
You should only respond in format as described below (a Python list of queries). PLEASE STRICTLY FOLLOW THE FORMAT. DO NOT RETURN ANYTHING ELSE. START YOUR RESPONSE WITH '['.
[response format]: ['query1', 'query2']

Here are three examples:
claim: The CEO of twitter is Bill Gates.
response: ["Who is the CEO of twitter?", "CEO Twitter"]

claim: Michael Phelps is the most decorated Olympian of all time.
response: ["Who is the most decorated Olympian of all time?", "Michael Phelps"]

claim: ChatGPT is created by Google.
response: ["Who created ChatGPT?", "ChatGPT"]

Now complete the following(ONLY RESPONSE IN A LIST FORMAT, DO NOT RETURN OTHER WORDS!!! START YOUR RESPONSE WITH '[' AND END WITH ']'):
claim: {input}
response: 
"""
}

VERIFICATION_PROMPT = {
    "system": "You are a brilliant assistant.",
    "user": """You are given a piece of text. Your task is to identify whether there are any factual errors within the text.
When you are judging the factuality of the given text, you could reference the provided evidences if needed. The provided evidences may be helpful. Some evidences may contradict to each other. You must be careful when using the evidences to judge the factuality of the given text.
The response should be a dictionary with three keys - "reasoning", "factuality", "error", and "correction", which correspond to the reasoning, whether the given text is factual or not (Boolean - True or False), the factual error present in the text, and the corrected text.
The following is the given text
[text]: {claim}
The following is the provided evidences
[evidences]: {evidence}
You should only respond in format as described below. DO NOT RETURN ANYTHING ELSE. START YOUR RESPONSE WITH '{{'.
[response format]: 
{{
    "reasoning": "Why is the given text factual or non-factual? Be careful when you said something is non-factual. When you said something is non-factual, you must provide multiple evidences to support your decision.",
    "error": "None if the text is factual; otherwise, describe the error.",
    "correction": "The corrected text if there is an error.",
    "factuality": True if the given text is factual, False otherwise.
}}
"""
}



from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .factool_utils.chat_api import OpenAIChat
from .factool_utils.prompt import CLAIM_EXTRACTION_PROMPT


@Solver.register("factool_claimprocessor", "response", "claims")
class FactoolClaimProcessor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.gpt_model = self.global_config.get("factool_gpt_model", "gpt-4o")
        self.gpt = OpenAIChat(self.gpt_model)
        self.claim_prompt = CLAIM_EXTRACTION_PROMPT

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        response = state.get(self.input_name)

        claims = self._claim_extraction(responses=[response])[0]

        extracted_claims = [claim["claim"] for claim in claims]

        state.set(self.output_name, extracted_claims)
        return True, state

    def _claim_extraction(self, responses):
        messages_list = [
            [
                {"role": "system", "content": self.claim_prompt["system"]},
                {
                    "role": "user",
                    "content": self.claim_prompt["user"].format(input=response),
                },
            ]
            for response in responses
        ]
        return self.gpt.run(messages_list, list)



import os
import subprocess
from pathlib import Path
from jinja2 import Environment, FileSystemLoader

from importlib import resources as pkg_resources

from openfactcheck.utils.logging import get_logger
from openfactcheck.templates import report as report_dir

# Get the logger
logger = get_logger(__name__)

# Import latex template
report_template_path = str(pkg_resources.files(report_dir) / "llm_evaluation_report.tex")

def create_latex_report(model_name, report_path):
    """
    Fill data in tex templates.
    """

    loader = FileSystemLoader(os.path.dirname(report_template_path))
    env = Environment(loader=loader)
    data = {
        "model_name": model_name.replace("_", " "),
        "snowballing_barplot": "snowballing_barplot.png",
        "snowballing_cm": "snowballing_cm.png",
        "selfaware_barplot": "selfaware_barplot.png",
        "selfaware_cm": "selfaware_cm.png",
        "freshqa_barplot": "freshqa_barplot.png",
        "freetext_barplot": "freetext_barplot.png",
    }
    template = env.get_template(os.path.basename(report_template_path))
    latex = template.render(data)
    with open(Path(report_path) / ("main.tex"), "w", encoding="utf-8") as f:
        f.write(latex)

    return None


def compile_pdf(report_path):
    """
    Compile the latex file to pdf.
    """

    # Change the directory to the report path
    original_directory = os.getcwd()
    os.chdir(report_path)

    try:
        try:
            # Compile the latex file
            subprocess.run(["pdflatex", "main.tex"], timeout=60)
        except subprocess.TimeoutExpired:
            logger.error("Compilation of the report timed out.")
            raise Exception("Compilation of the report timed out.")
        

        # Rename the pdf file
        Path("main.pdf").replace("report.pdf")

        # Remove the auxiliary files
        for file in Path(".").glob("main*"):
            file.unlink()
        
        # Change the directory back to the original
        os.chdir(original_directory)

    except Exception as e:
        logger.error(f"Error compiling the report: {e}")
        raise Exception(f"Error compiling the report: {e}")


def create_report(model_name, report_path):
    """
    Create a pdf report.
    """

    logger.debug("Creating the report...")
    create_latex_report(model_name, report_path)

    logger.debug("Compiling the report...")
    compile_pdf(report_path)

    return Path(report_path) / "report.pdf"


def detect_streamlit() -> bool:
    try:
        from streamlit.runtime.scriptrunner import get_script_run_ctx

        if get_script_run_ctx() is not None:
            return True
        else:
            return False
        
    except ImportError:
        return False
    
    except Exception:
        return False
    


import os
import re
import torch
import string
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from transformers import AutoTokenizer, AutoModel
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.metrics import classification_report, confusion_matrix

os.environ["TOKENIZERS_PARALLELISM"] = "false"

class SelfAwareEvaluator():
    def __init__(self):
        pass

    def remove_punctuation(self, input_string):
        """
        Remove the punctuation from the input string.
        """
        input_string = input_string.strip().lower()
        if input_string and input_string[-1] in string.punctuation:
            return input_string[:-1]
        return input_string


    def cut_sentences(self, content):
        """
        Cut the content into sentences.
        """
        sentences = re.split(r"(\.|\!|\?|。|！|？|\.{6})", content)
        return sentences


    def cut_sub_string(self, input_string, window_size=5, punctuation=".,?!"):
        """
        Cut the input string into sub-strings of a fixed window size.
        """
        input_string = input_string.strip().lower()

        # If the input string is empty, return an empty list
        if len(input_string) < 2:
            return [""]
        
        # Remove the punctuation
        if input_string[-1] in punctuation:
            input_string = input_string[:-1]

        # Split the string into words
        string_list = input_string.split()

        # If the length of the string is less than the window size, return the string
        length = len(string_list)
        if length <= window_size:
            return [input_string]
        else:
            res = []
            for i in range(length - window_size + 1):
                sub_string = " ".join(string_list[i: i + window_size])
                if sub_string != "" or sub_string != " ":
                    res.append(sub_string)
            return res


    def group_cosine_similarity(self, model, tokenizer, sentences1, sentences2):
        """
        Calculate the cosine similarity between two groups of sentences.
        """
        # Move the model to the device
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        model.to(device)
        
        # Encode the sentences
        encoded1 = tokenizer(sentences1, padding=True, truncation=True, return_tensors="pt").to(device)
        encoded2 = tokenizer(sentences2, padding=True, truncation=True, return_tensors="pt").to(device)

        # Get the embeddings
        embeddings1 = model(**encoded1).pooler_output
        embeddings2 = model(**encoded2).pooler_output

        # Calculate the cosine similarity
        similarities = cosine_similarity(embeddings1.detach().cpu().numpy(), embeddings2.detach().cpu().numpy())

        return similarities


    def get_unanswerable(self, response, model, tokenizer, threshold=0.75):
        """
        Predict whether the response is unanswerable or not.
        """
        # List of uncertain words
        uncertain_list = [
            "The answer is unknown.",
            "The answer is uncertain.",
            "The answer is unclear.",
            "There is no scientific evidence.",
            "There is no definitive answer.",
            "There is no right answer.",
            "There is much debate.",
            "There is no known case.",
            "There is no concrete answer to this question.",
            "There is no public information available.",
            "It is impossible to know.",
            "It is impossible to answer.",
            "It is difficult to predict.",
            "It is not known.",
            "We do not know.",
            "I'm not sure.",
        ]
        uncertain_list = [self.remove_punctuation(_) for _ in uncertain_list]

        # Check if the response contains uncertain words
        pred_unanswerable = False
        response = response.lower()
        for uncertain in uncertain_list:
            if uncertain in response:
                pred_unanswerable = True

        # If the response does not contain uncertain words, check the cosine similarity
        if pred_unanswerable == False:
            # Cut the response into sentences
            sub_sen_list = self.cut_sentences(response)

            # Cut the sentences into sub-strings
            sub_str_list = []
            for sub_sen in sub_sen_list:
                if len(sub_sen) >= 2:
                    sub_str_list.extend(self.cut_sub_string(sub_sen))
            
            # Calculate the cosine similarity
            if len(sub_str_list) != 0:
                similarities = self.group_cosine_similarity(model, tokenizer, sub_str_list, uncertain_list)
            else:
                similarities = [0]

            # Check if the maximum similarity is greater than the threshold
            max_uncertainty = np.max(similarities)

            # If the maximum similarity is greater than the threshold, predict unanswerable
            if max_uncertainty > threshold:
                pred_unanswerable = True

        return pred_unanswerable
    
    def selfaware_barplot(self, result: dict, fig_path: str = "", save: bool = False):
        """
        Create a bar plot of the performance on the SelfAware dataset.

        Parameters
        ----------
        result : dict
            The evaluation results for the LLM responses on the SelfAware dataset.
        fig_path : str
            The path to save the figure.
        save : bool, optional
            Whether to save the figure, by default True.
        """

        # Data
        unanswerable_as_pos = result["unanswerable_as_pos"]
        answerable_as_pos = result["answerable_as_pos"]

        # Remove support
        unanswerable_as_pos.pop("support", None)
        answerable_as_pos.pop("support", None)

        # Extract the accuracy values for each topic
        metrics = list(unanswerable_as_pos.keys())
        unanswerable_values = [round(v, 2) for k, v in unanswerable_as_pos.items()]
        answerable_values = [round(v, 2) for k, v in answerable_as_pos.items()]

        # Create a new figure
        fig, ax = plt.subplots()

        # Number of groups
        n_groups = len(metrics)
        index = np.arange(n_groups)
        bar_width = 0.35

        # Select two colors from the "rocket" palette
        colors = sns.color_palette("rocket", n_colors=10) 
        color_unanswerable = colors[1]
        color_answerable = colors[7] 

        # Plotting both sets of data
        bars1 = ax.bar(index, unanswerable_values, bar_width, label='Unanswerable as Positive', color=color_unanswerable)
        bars2 = ax.bar(index + bar_width, answerable_values, bar_width, label='Answerable as Positive', color=color_answerable)

        # Adding values on top of each bar
        for bar in bars1:
            ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height(), f'{bar.get_height():.2f}',
                    ha='center', va='bottom', color='black', rotation='horizontal')
        for bar in bars2:
            ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height(), f'{bar.get_height():.2f}',
                    ha='center', va='bottom', color='black', rotation='horizontal')

        # Set x-axis labels
        ax.set_xticks(index + bar_width / 2)
        ax.set_xticklabels(metrics)

        # Set y-axis limits to accommodate annotations
        ax.set_ylim((0, max(unanswerable_values + answerable_values) + 0.1)) 

        # Adding labels and title
        ax.set_xlabel("Metrics")
        ax.set_ylabel("Performance")
        ax.set_title("Performance on SelfAware Dataset")
        ax.legend()

        if save:
            # Save the figure
            plt.tight_layout()
            plt.savefig(os.path.join(fig_path, "selfaware_barplot.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "selfaware_barplot.png"), format="png")

        # Return the figure
        return fig  
    
    def selfaware_cm(self, labels: list, preds: list, fig_path: str = "", save: bool = False):
        """
        Create a confusion matrix for the SelfAware dataset.

        Parameters
        ----------
        labels : list
            The true labels.
        preds : list
            The predicted labels.
        fig_path : str
            The path to save the figure.
        save : bool, optional
            Whether to save the figure, by default True.
        """

        # Create a new figure
        fig, ax = plt.subplots()

        # Compute confusion matrix
        cm = sns.heatmap(confusion_matrix(labels, preds), annot=True, fmt="d", cmap="Blues", ax=ax)

        # Adding labels and title
        plt.xticks(ticks=[0.5, 1.5], labels=["Answerable", "Unanswerable"])
        plt.yticks(ticks=[0.5, 1.5], labels=["Answerable", "Unanswerable"])
        plt.ylabel("True label")
        plt.xlabel("Predicted label")
        plt.title("Confusion Matrix on SelfAware dataset.")

        if save:
            # Save the figure
            plt.tight_layout()
            plt.savefig(os.path.join(fig_path, "selfaware_cm.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "selfaware_cm.png"), format="png")

        # Return the figure
        return fig     

    def evaluate_selfaware(self, llm_responses):
        # Load the model
        model_name = "princeton-nlp/sup-simcse-roberta-large"
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        model = AutoModel.from_pretrained(model_name).to(device)

        # Evaluate the LLM responses
        labels, preds = [], []
        for item in llm_responses:
            # gold label: whether the question is answerable or not.
            labels.append(item["label_unanswerable"])
            # identify whether the model response answered the question or not.
            preds.append(self.get_unanswerable(item["response"], model, tokenizer))

        # Change the key names
        result = classification_report(labels, preds, output_dict=True, zero_division=0)

        # Renaming keys based on the expected output dictionary structure
        # Unanswerable as positive class and answerable as negative class
        if "True" in result:
            result['unanswerable_as_pos'] = result.pop("True")
        if "False" in result:
            result['answerable_as_pos'] = result.pop('False')

        return result, labels, preds




from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .rarr_utils.agreement_gate import run_agreement_gate
from .rarr_utils.functional_prompt import AGREEMENT_GATE_PROMPT


@Solver.register("rarr_verifier", "claims_with_evidences", "label")
class RARRAgreementGate(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_evidences_per_question = args.get("max_evidences_per_question", 1)
        self.model = self.global_config.get("rarr_model", "gpt-4o-instruct")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims_with_evidences = state.get(self.input_name)
        results = []
        for claim, evidences in claims_with_evidences.items():
            result = {}
            evidences = evidences[: self.max_evidences_per_question]
            labels = []
            for query, evidence in evidences:
                gate = run_agreement_gate(
                    claim=claim,
                    context=None,
                    query=query,
                    evidence=evidence,
                    model=self.model,
                    prompt=AGREEMENT_GATE_PROMPT,
                )
                labels.append(gate["is_open"])
            result["claim"] = claim
            result["evidences"] = evidences
            result["labels"] = labels
            result["factuality"] = all(labels)
            results.append(result)
        state.set(self.output_name, all([x["factuality"] for x in results]))
        state.set("detail", results)
        return True, state



import os
import json
import uuid
import pandas as pd
from importlib import resources as pkg_resources

from openfactcheck import OpenFactCheck
from openfactcheck.utils.logging import get_logger
from openfactcheck.evaluator.llm.evaluate_snowballing import SnowballingEvaluator
from openfactcheck.evaluator.llm.evaluate_selfaware import SelfAwareEvaluator
from openfactcheck.evaluator.llm.evaluate_freshqa import FreshQAEvaluator
from openfactcheck.evaluator.llm.evaluate_freetext import FreeTextEvaluator
from openfactcheck.evaluator.llm.report import create_report

from openfactcheck import data as data_dir

# Get the logger
logger = get_logger(__name__)

# Import LLM Evaluation Dataset
default_dataset_path = str(pkg_resources.files(data_dir))
default_output_path = "tmp/output/llm_evaluator"

class LLMEvaluator(SnowballingEvaluator, SelfAwareEvaluator, FreshQAEvaluator, FreeTextEvaluator):
    """
    This class is used to evaluate the performance of a Language Model.

    Parameters
    ----------
    model_name : str
        The name of the Language Model.
    input_path : Union[str, pd.DataFrame]
        The path to the CSV file or the DataFrame containing the LLM responses.
        The CSV file should have the following two columns:
        - index: The index of the response.
        - response: The response generated by the LLM.
    output_path : str
        The path to store the output files.
    dataset_path : str
        The path to the dataset file containing the questions.
    datasets : list
        The list of datasets to evaluate the LLM on. 
    analyze : bool
        Whether to analyze the results.
    save_plots : bool
        Whether to save the plots.
    save_report : bool
        Whether to save the report.

    Attributes
    ----------
    model_name : str
        The name of the Language Model.
    run_id : str
        The unique identifier for the run.
    input_path : Union[str, pd.DataFrame]
        The path to the CSV file or the DataFrame containing the LLM responses.
    output_path : str
        The path to store the output files.
    dataset_path : str
        The path to the dataset file containing the questions.
    datasets : list
        The list of datasets to evaluate the LLM on.
    combined_result : dict
        The combined evaluation results for all datasets.

    Methods
    -------
    evaluate(model_name: str, input_path: Union[str, pd.DataFrame], output_path: str = "", dataset_path: str = "", datasets: list = ["snowballing"], analyze: bool = True, save_plots: bool = True, save_report: bool = True):
        This function evaluates the performance of the Language Model.
    read_input():
        This function reads the input file and dataset file and returns a DataFrame containing the combined data.
    filter_responses(df: pd.DataFrame, dataset: str):
        Filter the responses based on the dataset.
    generate_plots(fig_path, save_plots=True):
        Generate plots for the evaluation
    """
    def __init__(self, ofc: OpenFactCheck):
        SnowballingEvaluator.__init__(self)
        SelfAwareEvaluator.__init__(self)
        FreshQAEvaluator.__init__(self)
        FreeTextEvaluator.__init__(self, ofc)
        self.logger = logger

        # Set the attributes
        self.model_name = None
        self.run_id = str(uuid.uuid4().hex)
        self.input_path = None
        self.dataset_path = None
        self.output_path = None
        self.datasets = None

        self.combined_result = None
        self.labels = None
        self.predictions = None

        self.logger.info(f"LLM Evaluator initialized with run_id: {self.run_id}")

    def read_input(self):
        """
        This function reads the input file and dataset file and returns a DataFrame containing the combined data.
        """

        # Check if the input_path is a DataFrame
        if isinstance(self.input_path, pd.DataFrame):
            df_responses = self.input_path
        else:
            # Read the CSV file
            self.logger.info(f"Reading the LLM responses from {self.input_path}...")
            df_responses = pd.read_csv(self.input_path)

        # Check the number of columns and if any response is missing
        assert df_responses.shape[1] == 2, "The LLM responses should have 2 columns."

        # Use the first column as index and rename the index and response column
        df_responses.set_index(df_responses.columns[0], inplace=True)
        df_responses.index.name = None
        df_responses.columns = ["response"]

        # Read the avaliable datasets
        self.logger.info(f"Reading the dataset from {self.dataset_path}...")
        df_dataset = pd.DataFrame()
        # Loop through each file in the directory
        for filename in os.listdir(self.dataset_path):
            if filename.endswith('.jsonl'):  # Check if the file is a JSONL file
                file_path = os.path.join(self.dataset_path, filename)

                logger.info(f"Reading {filename.split('.')[0]} dataset...")
                # Read the JSONL file and append it to the combined DataFrame
                df = pd.read_json(file_path, lines=True)
                df_dataset = pd.concat([df_dataset, df], ignore_index=True)

        # Combine the responses and questions
        assert len(df_responses) == len(df_dataset), "The number of responses and questions should be the same."
        df_combined = pd.concat([df_dataset, df_responses], axis=1)

        return df_combined       

    @staticmethod
    def filter_responses(df: pd.DataFrame, dataset: str):
        logger.info(f"Filtering responses for dataset: {dataset}...")
        # Filter the DataFrame based on the 'source' column directly
        if dataset == "snowballing":
            # Filter the DataFrame based on the 'source' column directly
            filtered_df = df[df['source'] == dataset]
            
            # Create a new DataFrame with only the required columns
            responses_df = filtered_df[['topic', 'response']].copy()
            
            # Convert the DataFrame to a list of dictionaries if needed
            responses = responses_df.to_dict(orient='records')

        elif dataset == "selfaware":
            selfaware_subset = [
                "selfaware-hotpot_train",
                "selfaware-squadqa_train",
                "selfaware-triviaqa_train",
                "selfaware-squadqa_dev",
                "selfaware-hotpot_dev",
                "selfaware-triviaqa_dev",
                "selfaware-SelfAware",
            ]

            responses = []
            for k, row in df.iterrows():
                if row["source"] in selfaware_subset:
                    responses.append({
                        "label_unanswerable": row["ability_to_test"].lstrip("answerable: ") == "False",
                        "response": row["response"]})
                    
        elif dataset == "freshqa":
            responses = []
            for k, row in df.iterrows():
                if row["source"] == dataset:
                    responses.append(
                        {
                            "question": row["question"],
                            "reference_answer": row["reference_answer"],
                            "response": row["response"],
                        }
                    )
        
        elif dataset in ["factoolqa", "felm-wk", "factcheck-bench"]:
            responses = []
            for k, row in df.iterrows():
                if row["source"] == dataset:
                    responses.append(
                        {
                            "source": row["source"],
                            "prompt": row["prompt"],
                            "response": row["response"],
                        }
                    )

        elif dataset == "factscore-bio":
            factscore_subset = [
                "factscore-labelled",
                "factscore-unlabelled",
            ]

            responses = []
            for k, row in df.iterrows():
                if row["source"] in factscore_subset:
                    responses.append(
                        {
                            "source": row["source"],
                            "prompt": row["prompt"],
                            "response": row["response"],
                        }
                    )
        else:
            raise ValueError(f"Dataset {dataset} is not supported.")
        
        return responses
    
    def generate_plots(self, fig_path: str = "", save_plots=True):
        # Create a bar plot of the accuracy of the LLM responses on the Snowballing dataset
        # for each topic and the overall accuracy.
        plots = {}
        for dataset in self.combined_result:
            if dataset == "snowballing":
                plots["snowballing"] = {}
                plots["snowballing"]["barplot"] = self.snowballing_barplot(self.combined_result[dataset], fig_path, save=save_plots)
                plots["snowballing"]["cm"] = self.snowballing_cm(self.labels[dataset], self.predictions[dataset], fig_path, save=save_plots)

            elif dataset == "selfaware":
                plots["selfaware"] = {}
                plots["selfaware"]["barplot"] = self.selfaware_barplot(self.combined_result[dataset], fig_path, save=save_plots)
                plots["selfaware"]["cm"] = self.selfaware_cm(self.labels[dataset], self.predictions[dataset], fig_path, save=save_plots)
            
            elif dataset == "freshqa":
                plots["freshqa"] = {}
                plots["freshqa"]["piechart"] = self.freshqa_piechart(self.combined_result[dataset], fig_path, save=save_plots)

            elif dataset == "freetext":
                plots["freetext"] = {}
                plots["freetext"]["barplot"] = self.freetext_barplot(self.combined_result["freetext"], fig_path, save=save_plots)

        return plots
    
    def generate_report(self, report_path: str):
        # Create a LaTeX report and return the path to the generated PDF
        return create_report(self.model_name, report_path)

    def evaluate(self, 
                 model_name: str,
                 input_path: str, 
                 output_path: str = "",
                 dataset_path: str = "", 
                 datasets: list = [
                     "snowballing",
                     "selfaware",
                     "freshqa",
                     "factoolqa",
                     "felm-wk",
                     "factcheck-bench",
                     "factscore-bio"
                 ],
                 analyze: bool = True,
                 save_report: bool = True):
        self.logger.info("Evaluating LLM responses...")

        # Set the attributes
        self.model_name = model_name
        self.input_path = input_path
        self.output_path = output_path
        self.dataset_path = dataset_path
        self.datasets = datasets
        
        # Check if the output path is provided (if not, use the default template)
        if self.output_path == "":
            self.output_path = default_output_path

        # Check if the output path exists (if not, create it)
        if not os.path.exists(f"{self.output_path}/{self.run_id}"):
            os.makedirs(f"{self.output_path}/{self.run_id}")
        
        # Check if the questions path is provided (if not, use the default template)
        if self.dataset_path == "":
            self.dataset_path = default_dataset_path

        # Read the input
        self.logger.info("Reading the input...")
        df = self.read_input()
        self.logger.info(f"Combined data contains {len(df)} rows")
        
        # Evaluate model responses over each dataset
        self.combined_result = {}
        self.labels = {}
        self.predictions = {}
        for dataset in self.datasets:
            logger.info(f"Evaluating responses for dataset: {dataset}...")
            if dataset == "snowballing":
                # Filter responses based on the dataset
                responses = self.filter_responses(df, dataset)

                # Evaluate the responses
                result, labels, preds = self.evaluate_snowballing(responses)

                # Store the output and save the results
                df_out = pd.DataFrame({"gold_labels": labels, "predictions": preds})
                df_out.to_json(f"{self.output_path}/{self.run_id}/{dataset}_output.jsonl", orient="records", lines=True)
                self.combined_result[dataset] = result
                self.labels[dataset] = labels
                self.predictions[dataset] = preds
            
            elif dataset == "selfaware":
                # Filter responses based on the dataset
                responses = self.filter_responses(df, dataset)

                # Evaluate the responses
                result, labels, preds = self.evaluate_selfaware(responses[:30])

                # Store the output and save the results
                df_out = pd.DataFrame({"gold_labels": labels, "predictions": preds})
                df_out.to_json(f"{self.output_path}/{self.run_id}/{dataset}_output.jsonl", orient="records", lines=True)
                self.combined_result[dataset] = result
                self.labels[dataset] = labels
                self.predictions[dataset] = preds

            elif dataset == "freshqa":
                # Filter responses based on the dataset
                responses = self.filter_responses(df, dataset)

                # Evaluate the responses
                result, raw_evals, preds = self.evaluate_freshqa(responses[:30])

                # Store the output and save the results
                df_out = pd.DataFrame({"raw_evaluations": raw_evals, "predictions": preds})
                df_out.to_json(f"{self.output_path}/{self.run_id}/{dataset}_output.jsonl", orient="records", lines=True)
                self.combined_result[dataset] = result

            elif dataset in ["factoolqa", "felm-wk", "factcheck-bench", "factscore-bio"]:
                # Check if the freetext key exists
                if self.combined_result.get("freetext") is None:
                    self.combined_result["freetext"] = {}

                # Filter responses based on the dataset
                responses = self.filter_responses(df, dataset)

                # Evaluate the responses
                results, evaluations = self.evaluate_freetext(responses[:30], self.model_name, self.run_id)

                # Store the output and save the results
                df_out = pd.DataFrame(evaluations)
                df_out.to_json(f"{self.output_path}/{self.run_id}/{dataset}_output.jsonl", orient="records", lines=True)
                self.combined_result["freetext"][dataset] = results
            
            else:
                logger.error(f"Dataset {dataset} is not supported.")
                raise ValueError(f"Dataset {dataset} is not supported.")

            logger.info(f"Finished evaluating responses for dataset: {dataset}")

        # save all evaluation results
        with open(f"{self.output_path}/{self.run_id}/result.json", "w") as json_file:
            json.dump(self.combined_result, json_file, indent=4)

        # Analyze the results
        if analyze:
            self.logger.info("Analyzing the results...")
            self.generate_plots(save_plots=True, fig_path=f"{self.output_path}/{self.run_id}")
        
        # Create a report
        if save_report:
            self.logger.info("Creating the report...")
            self.generate_report(report_path=f"{self.output_path}/{self.run_id}")
        
        return self.combined_result
                

        


        


import json
from typing import Any

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

from .factcheckgpt_utils.prompt import VERIFY_PROMPT
from .factcheckgpt_utils.openai_api import gpt
from .factcheckgpt_utils.data_util import save_to_file
from .factcheckgpt_utils.prompt import IDENTIFY_STANCE_PROMPT, IDENTIFY_STANCE_PROMPT_FUNC
from .factcheckgpt_utils.nli import nli_infer


@Solver.register("factcheckgpt_verifier", "claims_with_evidences", "label")
class FactCheckGPTVerifier(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.stance_model = args.get("stance_model", "gpt-4o")
        self.num_retries = self.global_config.get("num_retries", 3)
        # self.system_role = args.get("system_role", "You are a helpful factchecker assistant.")
        self.system_role = "You are a helpful factchecker assistant."
        self.verify_retries = args.get("verify_retries", 3)
        self.stance_map = {1: "support", -1: "refute", 0: "irrelevant"}

    def verify_by_stance(
        self,
        claim: str,
        evidences: list[str],
    ) -> Any:
        labels = []
        for evidence in evidences:
            labels.append(self.stance(evidence, claim))

        # based on stances of evidence, determine the true/false claim by rules
        # if there is one evidence supports, we assume it is correct
        if 1 in labels:
            return 1
        # if there isn't support, but refute and irrelevant, we regard as false
        elif -1 in labels:
            return -1
        else:
            # all irrelevant
            return 0

    def identify_stance_gpt(self, evidence, claim):
        user_input = IDENTIFY_STANCE_PROMPT_FUNC.format(claim=claim, evidence=evidence)
        r = gpt(user_input, model=self.stance_model, system_role=self.system_role, num_retries=self.num_retries)
        label = 0
        try:
            label = eval(r)
        except Exception as e:
            print(f"An unexpected error occurred: {e}.")
        return label

    def stance(self, evidence, claim, model="gpt-4o"):
        """input: a claim and an evidence
        output: label in [support, refute, irrelevant]"""
        label = 0
        if self.stance_model == "nli":
            label = nli_infer(premise=evidence, hypothesis=claim)
        elif "gpt" in self.stance_model:
            label = self.identify_stance_gpt(evidence, claim)
        else:
            print("Check the model argument, choose either gpt or nli model")
        return label

    def verify_claim(self, claim: str, evidences: list[str]) -> dict[str, Any]:
        results = None
        user_input = VERIFY_PROMPT.format(claim=claim, evidence=evidences)
        r = ""
        for _ in range(self.verify_retries):
            r = gpt(
                user_input,
                model=self.stance_model,
                system_role=self.system_role,
                num_retries=self.num_retries,
            )
            try:
                results = eval(r)
                break
            except Exception as e:
                try:
                    results = json.loads(r)
                except Exception as e:
                    print(f"An unexpected error occurred to parse json {r}: {e}.")
                    save_to_file(r, "verification_error.txt")
                    print(f"An unexpected error occurred to eval {r}: {e}.")

        if isinstance(results, dict):
            return results
        else:
            print(f"Error output {r}. It does not output a dict, return factual label by stance aggregation.")
            factual_label = self.verify_by_stance(claim, evidences)
            results = {"reasoning": "", "error": "", "correction": "", "factuality": factual_label}
            return results

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims_with_evidences = state.get(self.input_name)
        results = []
        for claim, evidences in claims_with_evidences.items():
            result = self.verify_claim(claim, [x[1] for x in evidences])
            result["claim"] = claim
            result["evidences"] = evidences
            results.append(result)
        state.set(self.output_name, all([x["factuality"] > 0 for x in results]))
        state.set("detail", results)
        return True, state



from typing import Any, Optional

from openfactcheck.utils.logging import get_logger

# Get the logger
logger = get_logger(__name__)


class FactCheckerState:
    """
    A class to manage the state of a fact-checking system.

    It holds a question and its corresponding response, and provides methods
    to set and get these attributes dynamically.
    """

    def __init__(self, question: Optional[str] = None, response: Optional[str] = None) -> None:
        """
        Initialize the FactCheckerState object.

        Parameters
        ----------
        question : Optional[str]
            The question to be fact-checked.
        response : Optional[str]
            The response to the question.
        """
        self.question: Optional[str] = question
        self.response: Optional[str] = response

    def set(self, name: str, value: Any) -> None:
        """
        Set an attribute of the state object.

        Parameters
        ----------
        name : str
            The name of the attribute to set.
        value : Any
            The value to set for the attribute.
        """
        if hasattr(self, name):
            logger.warning(f"Modifying existing attribute '{name}'")
        setattr(self, name, value)

    def get(self, name: str) -> Any:
        """
        Get an attribute of the state object.

        Parameters
        ----------
        name : str
            The name of the attribute to retrieve.

        Returns
        -------
        Any
            The value of the requested attribute.

        Raises
        ------
        ValueError
            If the attribute does not exist.
        """
        if not hasattr(self, name):
            raise ValueError(f"Attribute '{name}' does not exist")
        return getattr(self, name)

    def __str__(self) -> str:
        """
        Return a string representation of the state object.

        Returns
        -------
        str
            A string representation of the object's dictionary.
        """
        return str(self.__dict__)

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of the state object.

        Returns
        -------
        Dict[str, Any]
            A dictionary containing the object's attributes.
        """
        return self.__dict__



import csv
import json
import numpy as np
from collections import Counter
from typing import Dict, List, Any


def save_to_file(text, filename='error_output.txt'):
    """Save a string to a file line by line."""
    with open(filename, 'a', encoding='utf-8') as file:
        file.write(text + '\n')


def majority_vote(input_list):
    # Use Counter to count occurrences of each element
    counter = Counter(input_list)

    # Find the element with the maximum count (majority)
    majority_element = max(counter, key=counter.get)

    # Return the majority element
    return majority_element


def is_float(string):
    if string.replace(".", "").isnumeric():
        return True
    else:
        return False


def save_json(dictionary: Dict[str, Any], save_dir: str) -> None:
    # Serializing json
    json_object = json.dumps(dictionary, indent=4, ensure_ascii=False)

    # Writing to sample.json
    with open(save_dir, "w", encoding='utf-8') as outfile:
        outfile.write(json_object)


def read_json(filepath: str) -> Dict[str, Any]:
    data = {}
    with open(filepath, 'r', encoding='utf-8') as file:
        data = json.load(file)
    return data


def list_to_dict(data: List[Dict[str, Any]]) -> Dict[int, Any]:
    temp = {}
    for i, d in enumerate(data):
        temp[i] = d
    return temp


def load_jsonl(path):
    data = []
    with open(path, 'r', encoding='utf-8') as reader:
        for line in reader:
            data.append(json.loads(line))
    return data


# def load_jsonl(input_path) -> list:
#     """
#     Read list of objects from a JSON lines file.
#     """
#     data = []
#     with open(input_path, 'r', encoding='utf-8') as f:
#         for line in f:
#             data.append(json.loads(line.rstrip('\n|\r')))
#     print('Loaded {} records from {}'.format(len(data), input_path))
#     return data

def dump_jsonl(data, output_path, append=False):
    """
    Write list of objects to a JSON lines file.
    """
    mode = 'a+' if append else 'w'
    with open(output_path, mode, encoding='utf-8') as f:
        for line in data:
            json_record = json.dumps(line, ensure_ascii=False)
            f.write(json_record + '\n')
    print('Wrote {} records to {}'.format(len(data), output_path))


def cosine(u, v):
    """based on embeddings and calculate cosine similarity"""
    return np.dot(u, v) / (np.linalg.norm(u) * np.linalg.norm(v))


def read_csv(input_file, quotechar=None):
    with open(input_file, "r", encoding="utf-8") as f:
        reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
        lines = []
        for line in reader:
            lines.append(line)
    return lines


def save_csv(header, data, output_file):
    with open(output_file, 'w', encoding='UTF8', newline='') as f:
        writer = csv.writer(f, delimiter='\t')
        # write the header
        writer.writerow(header)
        # write multiple rows
        writer.writerows(data)


def save_array(filename, embeddings):
    # save embeddings into file
    with open(filename, 'wb') as f:
        np.save(f, embeddings)


def load_array(filename):
    with open(filename, 'rb') as f:
        a = np.load(f)
    return a


def read_txt(input_file):
    with open(input_file, "r", encoding="utf-8") as f:
        return f.readlines()


def save_txt(data, output_file):
    with open(output_file, "w", encoding="utf-8") as writer:
        writer.write("\n".join(data))


def clean_text(text):
    for mark in ['"', '-', '\t', ' ']:
        for i in [5, 4, 3, 2]:
            marks = mark * i
            text = text.replace(marks, '')
    return text



"""All prompts used for DFC prompting."""

DECONTEXTILISATION_PROMPT = """Decompose and decontextualise a document into independently meaningful sentences. This process will make each sentence stand alone that can be verified independently.

Input: Mary is a five-year old girl. She likes playing piano. She doesn't like cookies. 
Output: 
Mary is a five-year old girl.
Mary likes playing piano.
Mary doesn't like cookies. 

Input: Google began as an online search firm, but it now offers more than 50 Internet services and products, from e-mail and online document creation to software for mobile phones and tablet computers. In addition, its 2012 acquisition of Motorola Mobility put it in the position to sell hardware in the form of mobile phones. 
Ouput: 
Google began as an online search firm.
Google now offers more than 50 Internet services and products.
Google offers from e-mail and online document creation to software for mobile phones and tablet computers.
Google 2012 acquisition of Motorola Mobility put it in the position to sell hardware in the form of mobile phones.

Input: """


CHECK_WORTHINESS_LABEL_ONLY_PROMPT = """Identify whether this claim is an opinion or factual, and whether it is checkworthy or not in the context of fact-checking. Just return two labels without explanation.
I think Apple is a good company.
opinon, not checkworthy
Preslav is a professor in MBZUAI.
factual, checkworthy
Friends is a great TV series.
opinion, not checkworthy
The Stanford Prison Experiment was conducted in the basement of Encina Hall.
factual, checkworthy
"""


ENTITY_EXTRACTION_PROMPT = """Extract all entities of a claim.
Input: Google now offers more than 50 Internet services and products.
Output: Google, Internet services, product
Input: Donald John Trump is an American politician, media personality, and businessman.
Output: Donald John Trump, American politician, media personality, businessman
Input: """       


QGEN_PROMPT = """Give a list of queries using for searching related information for a claim. 
Input: Google now offers more than 50 Internet services and products. 
Output: What does Google offers now? 
How many service and product does Google offer? 
Google, more than 50 Internet services, products 
Input: Donald John Trump is an American politician, media personality, and businessman.
Output: Who is Donald John Trump?
Give information of Donald John Trump. 
Donald John Trump, American politician 
Donald John Trump, media personality 
Donald John Trump, businessman
Input: """


STANCE_DETECTION_PROMPT = """Determine whether the evidence support the claim or not. Choose label from [support, partial support, refute, other] and explain why. 
Support means we can entail the claim by the evidence.
Partial support means: part of the information presented in the claim appear in the evidence.
Refute means that the evidence mention the same event as the claim, but a clear opposite fact. It should be highlighed that under refute, the evidence mentions the fact in the claim, they are closely relevant, but opposite meaning or stance.
Other means the evidence does not mention anything about the fact described in the claim, such that it neither supports nor refutes the claim.

Claim: Elon Musk is the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: support, statement 'he is also the founder, CEO and chief engineer of SpaceX' in evidence above supports the claim.

Claim: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the founder, CEO and chief engineer of SpaceX.
Stance: partial support.

Claim: Steve Jobs is the founder, CEO and chief engineer of SpaceX.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: refute.

Claim: Elon Musk is a professor in The Stanford University.
Evidence: Elon Musk is the owner and CEO of Twitter, and he is also the founder, CEO and chief engineer of SpaceX.
Stance: other, according to the evidence, I cannot judge whether the claim is true or not, not enough information, the evidence neither supports nor refutes.

Claim: On January 6, 2021, a mob of supporters of former President Donald Trump stormed the U.S. Capitol in an attempt to overturn the 2020 presidential election.
Evidence: On January 6, 2021, following the defeat of U.S. President Donald Trump in the 2020 presidential election, a mob of his supporters attacked the United States Capitol Building in Washington, D.C. The mob sought to keep Trump in power by preventing a joint session of Congress from counting the electoral college votes to formalize the victory of President-elect Joe Biden.
Stance: support.

Claim: The 2021 Capitol Hill riots resulted in the deaths of five people, including a Capitol police officer. 
Evidence: Five people died either shortly before, during, or following the riot: one was shot by Capitol Police, another died of a drug overdose, and three died of natural causes.
Stance: partial support, the evidence supports that fact that five deaths, but not sure whether they include a Capitol police officer or not.

Claim: More than 300 people have been charged with crimes related to the riots. 
Evidence: As of November 10, 2022, over 940 people had been charged in the Capitol breach. 
Stance: refute, evidence and claim are describing the same thing, the number of people who was charged is over 940, while more than 300 in the claim, so the evidence refutes the claim.

Claim: More than 300 people have been charged with crimes related to the riots. 
Evidence: The laptop computer taken from Pelosi's office was taken by 22-year-old Capitol rioter Riley Williams. Williams was arrested and indicted on eight counts, including theft of government property, obstructing an official proceeding, and assaulting or resisting police.
Stance: other, the evidence demonstrates something relevent to the fact in the claim, but it does not support or refute any information of it. 

Claim: {}
Evidence: {}
Stance: """


EDITOR_PROMPT = """Fix the claim according to the evidence.

Claim: Your nose switches back and forth between nostrils. When you sleep, you switch about every 45 minutes. This is to prevent a buildup of mucus. It’s called the nasal cycle.
Evidence: Although we don’t usually notice it, during the nasal cycle one nostril becomes congested and thus contributes less to airflow, while the other becomes decongested. On average, the congestion pattern switches about every 2 hours, according to a small 2016 study published in the journal PLOS One.
This suggests 45 minutes switch time in your statement is wrong.
Fix: Your nose switches back and forth between nostrils. When you sleep, you switch about every 2 hours. This is to prevent a buildup of mucus. It’s called the nasal cycle.

Claim: In the battles of Lexington and Concord, the British side was led by General Thomas Hall.
Evidence: Interesting Facts about the Battles of Lexington and Concord. The British were led by Lieutenant Colonel Francis Smith. There were 700 British regulars.
This suggests General Thomas Hall in your statement is wrong.
Fix: In the battles of Lexington and Concord, the British side was led by Lieutenant Colonel Francis Smith.

Claim: The Stanford Prison Experiment was conducted in the basement of Encina Hall, Stanford’s psychology building.
Evidence: Carried out August 15-21, 1971 in the basement of Jordan Hall, the Stanford Prison Experiment set out to examine the psychological effects of authority and powerlessness in a prison environment.
This suggests Encina Hall in your statement is wrong.
Fix: The Stanford Prison Experiment was conducted in the basement of Jordan Hall, Stanford’s psychology building.

Claim: The Havel-Hakimi algorithm is an algorithm for converting the adjacency matrix of a graph into its adjacency list. It is named after Vaclav Havel and Samih Hakimi.
Evidence: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. This construction is based on a recursive algorithm. The algorithm was published by Havel (1955), and later by Hakimi (1962).
This suggests the Havel-Hakimi algorithm’s functionality in your statement is wrong.
Fix: The Havel-Hakimi algorithm constructs a special solution if a simple graph for the given degree sequence exists, or proves that one cannot find a positive answer. It is named after Vaclav Havel and Samih Hakimi.

Claim: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Phil Ramone.
Evidence: On September 8, 2010, the original demo of this song, along with a remix by producer Michael Lloyd , was released as digital files in an effort to raise money for the Patrick Swayze Pancreas Cancer Resarch Foundation at Stanford University.
This suggests "Time of My Life" producer name in your statement is wrong.
Fix: "Time of My Life" is a song by American singer-songwriter Bill Medley from the soundtrack of the 1987 film Dirty Dancing. The song was produced by Michael Lloyd.

Claim: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 1.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.
Evidence: Phoenix Market City was opened in January 2013 and has the distinction of being the largest mall in the city of Pune, with the area of 3.4 million square feet. It is located in the Viman Nagar area of Pune.
This suggests the 1.4 million square feet of built-up space in your statment is wrong.
Fix: Phoenix Market City Pune is located on 21 acres of prime property in Pune. It is spread across four levels with approximately 3.4 million square feet of built-up space. The mall is owned and operated by Phoenix Mills Limited.

Claim: {claim}
Evidence: {evidence}
This suggests
""".strip()



from argparse import Namespace

from .utils.prompt_base import CHECK_WORTHINESS_LABEL_ONLY_PROMPT
from .utils.api import chatgpt

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("chatgpt_worthiness_filter", "claims", "claims")
class ChatGPTWorthinessFilter(StandardTaskSolver):
    def __init__(self, args: Namespace):
        super().__init__(args)

    # string to format labels
    def convert_checkworthy_output_to_labels(self, label: str) -> bool:
        # factual_labels, checkworthy_labels = [], []
        # for label in labels:
        #
        #     factual_labels.append(opinion_vs_factual)
        #     checkworthy_labels.append(checkworthy)
        #
        # print(factual_labels)
        # print(checkworthy_labels)
        label = label.lower()
        if label[-1] == ".":
            label = label[:-1]
        opinion_vs_factual, checkworthy = label.split(",")
        if "fact" in opinion_vs_factual:
            opinion_vs_factual = "factual"
        else:
            opinion_vs_factual = "opinion"

        if "not" in checkworthy or opinion_vs_factual == "opinion":
            checkworthy = False
        else:
            checkworthy = True
        return checkworthy

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        valid_claims = []
        for claim in claims:
            response = chatgpt(CHECK_WORTHINESS_LABEL_ONLY_PROMPT + claim)
            if self.convert_checkworthy_output_to_labels(response):
                valid_claims.append(claim)

        state.set(self.output_name, valid_claims)
        return True, state



import streamlit as st

def metric_card(
        background_color: str = "#FFF",
        border_size_px: int = 1,
        border_color: str = "#CCC",
        border_radius_px: int = 5,
        border_left_color: str = "#9AD8E1",
        label: str = "Label",
        value: str = "Value",
    ) -> None:
        
    html_str = f"""
    <div>
        <p style='font-size: 14px;
                  color: rgb(49, 51, 63);
                  height: auto;
                  min-height: 1.5rem;
                  vertical-align: middle;
                  flex-direction: row;
                  -webkit-box-align: center;
                  align-items: center;
                  margin-bottom: 0px;
                  display: grid;
                  background-color: {background_color};
                  border: {border_size_px}px solid {border_color};
                  padding: 5% 5% 5% 10%;
                  border-radius: {border_radius_px}px;
                  border-left: 0.5rem solid {border_left_color};'>
            {label}
            <br>
            <span style='font-size: 2.25rem;
                         padding-bottom: 0.25rem;'>
                {value}
            </span>
        </p>
    </div>
    """
    st.markdown(html_str, unsafe_allow_html=True)
    st.markdown('######')


import argparse

from openfactcheck import OpenFactCheck
from openfactcheck.lib import OpenFactCheckConfig

def parse_args():
    parser = argparse.ArgumentParser(description='Initialize OpenFactCheck with custom configuration.')
    
    # Add arguments here, example:
    parser.add_argument("--config-path", 
                        type=str, 
                        help="Config File Path",
                        default="config.json")
    
    # Parse arguments from command line
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    args = parse_args()

    def callback(index, sample_name, solver_name, input_name, output_name, input, output, continue_run):
        print(f"Callback: {index}, {sample_name}, {solver_name}, {input_name}, {output_name}, {input}, {output}, {continue_run}")

    config = OpenFactCheckConfig(args.config_path)
    results = OpenFactCheck(config).ResponseEvaluator.evaluate("Abraham Lincoln was the first president of the United States.",
                                                               callback=callback)


class Error(Exception):
    """Base class for other exceptions"""

class ConfigError(Error):
    """Raised when there is an error with the configurations"""

class ConfigValidationError(Error):
    """Raised when configuration validation fails"""


"""Utils for generating fake evidence given a query."""
import os
import time
from typing import Dict

import openai

openai.api_key = os.getenv("OPENAI_API_KEY")


def run_evidence_hallucination(
    query: str,
    model: str,
    prompt: str,
    num_retries: int = 5,
) -> Dict[str, str]:
    """Generates a fake piece of evidence via LLM given the question.

    Args:
        query: Query to guide the validity check.
        model: Name of the OpenAI GPT-3 model to use.
        prompt: The prompt template to query GPT-3 with.
        num_retries: Number of times to retry OpenAI call in the event of an API failure.
    Returns:
        output: A potentially inaccurate piece of evidence.
    """
    gpt3_input = prompt.format(query=query).strip()
    for _ in range(num_retries):
        try:
            response = openai.Completion.create(
                model=model,
                prompt=gpt3_input,
                temperature=0.0,
                max_tokens=256,
                stop=["\n", "\n\n"],
            )
            break
        except openai.error.OpenAIError as exception:
            print(f"{exception}. Retrying...")
            time.sleep(2)

    hallucinated_evidence = response.choices[0].text.strip()
    output = {"text": hallucinated_evidence, "query": query}
    return output



import os
import sys
import threading
import logging
from logging import (
    CRITICAL,  # noqa
    DEBUG,  # noqa
    ERROR,  # noqa
    FATAL,  # noqa
    INFO,  # noqa
    NOTSET,  # noqa
    WARN,  # noqa
    WARNING,  # noqa
)
from logging import captureWarnings as _captureWarnings
from typing import Optional

import datasets
import transformers

_lock = threading.Lock()
_default_handler: Optional[logging.Handler] = None

log_levels = {
    "debug": DEBUG,
    "info": INFO,
    "warning": WARNING,
    "error": ERROR,
    "critical": CRITICAL,
}

_default_log_level = logging.WARNING


class CustomStreamLoggingFormatter(logging.Formatter):
    """
    Custom log formatter class to colorize log messages based on their level.
    """

    # Define the color codes
    grey = "\x1b[38;20m"
    green = "\x1b[32;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    log_format = "%(levelname)s - %(module)s.%(funcName)s"

    # Define the log message formats for different log levels
    FORMATS = {
        logging.DEBUG: grey + log_format + reset,
        logging.INFO: green + log_format + reset,
        logging.WARNING: yellow + log_format + reset,
        logging.ERROR: red + log_format + reset,
        logging.CRITICAL: bold_red + log_format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno, self.log_format)
        formatter = logging.Formatter(log_fmt + ": %(message)s")
        return formatter.format(record)


class CustomFileLoggingFormatter(logging.Formatter):
    """
    Custom log formatter class for file logging.
    """

    log_format = "%(asctime)s - %(levelname)s - %(message)s"

    def format(self, record):
        formatter = logging.Formatter(self.log_format)
        return formatter.format(record)


def _get_default_logging_level():
    """
    If OPENFACTCHECK_VERBOSITY env var is set to one of the valid choices return that as the new default level.
    If it is not - fall back to `_default_log_level`
    """
    env_level_str = os.getenv("OPENFACTCHECK_VERBOSITY", None)
    if env_level_str:
        if env_level_str in log_levels:
            return log_levels[env_level_str]
        else:
            logging.getLogger().warning(
                f"Unknown option OPENFACTCHECK_VERBOSITY={env_level_str}, "
                f"has to be one of: { ', '.join(log_levels.keys()) }"
            )
    return _default_log_level


def _get_library_name() -> str:
    """
    Return the name of the library.
    """
    return __name__.split(".")[0]


def _get_library_root_logger() -> logging.Logger:
    """
    Return the root logger of the library.
    """
    return logging.getLogger(_get_library_name())


def _configure_library_root_logger() -> None:
    """
    Configure the library root logger with the default handler and formatter.
    """
    global _default_handler

    with _lock:
        if _default_handler:
            # This library has already configured the library root logger.
            return

        # Set sys.stderr as stream.
        _default_handler = logging.StreamHandler()
        _default_handler.setFormatter(CustomStreamLoggingFormatter())

        # set defaults based on https://github.com/pyinstaller/pyinstaller/issues/7334#issuecomment-1357447176
        if sys.stderr is None:
            sys.stderr = open(os.devnull, "w")

        # Flush the stderr stream to ensure that any logs are immediately written to the stream.
        _default_handler.flush = sys.stderr.flush

        # Apply our default configuration to the library root logger.
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(_default_handler)
        library_root_logger.setLevel(_get_default_logging_level())

        library_root_logger.propagate = False


def _configure_library_file_logger(file_name: str) -> None:
    global _default_handler

    with _lock:
        if _default_handler:
            # This library has already configured the library root logger.
            return

        # Set file_name as stream.
        _default_handler = logging.FileHandler(file_name)
        _default_handler.setFormatter(CustomFileLoggingFormatter())

        # set defaults based on
        if sys.stderr is None:
            sys.stderr = open(os.devnull, "w")

        # Flush the stderr stream to ensure that any logs are immediately written to the stream.
        _default_handler.flush = sys.stderr.flush

        # Apply our default configuration to the library root logger.
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(_default_handler)
        library_root_logger.setLevel(_get_default_logging_level())

        library_root_logger.propagate = False


def get_log_levels_dict():
    """
    Returns the dictionary of log levels used by OpenFactCheck.
    """
    return log_levels


def captureWarnings(capture):
    """
    Calls the `captureWarnings` method from the logging library to enable management of the warnings emitted by the
    `warnings` library.

    Read more about this method here:
    https://docs.python.org/3/library/logging.html#integration-with-the-warnings-module

    All warnings will be logged through the `py.warnings` logger.

    Careful: this method also adds a handler to this logger if it does not already have one, and updates the logging
    level of that logger to the library's root logger.
    """
    logger = get_logger("py.warnings")

    if not logger.handlers and _default_handler:
        logger.addHandler(_default_handler)

    logger.setLevel(_get_library_root_logger().level)

    _captureWarnings(capture)


def get_logger(
    name: Optional[str] = None,
    enable_file_logging: bool = False,
    file_name: str = "app.log",
) -> logging.Logger:
    """
    Return a logger with the specified name.

    This function is not supposed to be directly accessed unless you are writing a custom module.
    """

    # If the name is not provided, use the library name.
    if name is None:
        name = _get_library_name()

    # Configure the library root logger if it has not been configured yet.
    _configure_library_root_logger()

    if enable_file_logging:
        _configure_library_file_logger(file_name)

    # Return the logger with the specified name.
    return logging.getLogger(name)


def get_verbosity() -> int:
    """
    Return the current level for the OpenFactCheck's root logger as an int.

    Returns
    -------
    `int`
        The logging level of the OpenFactCheck's root logger.

    <Tip>

    OpenFactCheck has following logging levels:

    - 50: `openfactcheck.logging.CRITICAL` or `openfactcheck.logging.FATAL`
    - 40: `openfactcheck.logging.ERROR`
    - 30: `openfactcheck.logging.WARNING` or `openfactcheck.logging.WARN`
    - 20: `openfactcheck.logging.INFO`
    - 10: `openfactcheck.logging.DEBUG`

    </Tip>"""

    _configure_library_root_logger()
    return _get_library_root_logger().getEffectiveLevel()


def set_verbosity(verbosity: int | str) -> None:
    """
    Set the verbosity level for the OpenFactCheck's root logger.

    Args:
        verbosity (`int`):
            Logging level, e.g., one of:

            - `openfactcheck.logging.CRITICAL` or `openfactcheck.logging.FATAL`
            - `openfactcheck.logging.ERROR`
            - `openfactcheck.logging.WARNING` or `openfactcheck.logging.WARN`
            - `openfactcheck.logging.INFO`
            - `openfactcheck.logging.DEBUG`
    """

    _configure_library_root_logger()
    _get_library_root_logger().setLevel(verbosity)


def set_verbosity_info():
    """Set the verbosity to the `INFO` level."""
    return set_verbosity(INFO)


def set_verbosity_warning():
    """Set the verbosity to the `WARNING` level."""
    return set_verbosity(WARNING)


def set_verbosity_debug():
    """Set the verbosity to the `DEBUG` level."""
    return set_verbosity(DEBUG)


def set_verbosity_error():
    """Set the verbosity to the `ERROR` level."""
    return set_verbosity(ERROR)


def disable_default_handler() -> None:
    """Disable the default handler of the OpenFactCheck's root logger."""

    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().removeHandler(_default_handler)


def enable_default_handler() -> None:
    """Enable the default handler of the OpenFactCheck's root logger."""

    _configure_library_root_logger()

    assert _default_handler is not None
    _get_library_root_logger().addHandler(_default_handler)


def add_handler(handler: logging.Handler) -> None:
    """Adds a handler to the OpenFactCheck's root logger."""

    _configure_library_root_logger()

    assert handler is not None
    _get_library_root_logger().addHandler(handler)


def remove_handler(handler: logging.Handler) -> None:
    """Removes given handler from the OpenFactCheck's root logger."""

    _configure_library_root_logger()

    assert handler is not None and handler not in _get_library_root_logger().handlers
    _get_library_root_logger().removeHandler(handler)


def disable_propagation() -> None:
    """
    Disable propagation of the library log outputs. Note that log propagation is disabled by default.
    """

    _configure_library_root_logger()
    _get_library_root_logger().propagate = False


def enable_propagation() -> None:
    """
    Enable propagation of the library log outputs. Please disable the OpenFactCheck's default handler to
    prevent double logging if the root logger has been configured.
    """

    _configure_library_root_logger()
    _get_library_root_logger().propagate = True


# Disable Transformers and Datasets logging
transformers.logging.set_verbosity_error()
datasets.logging.set_verbosity_error()
logging.basicConfig(level=logging.ERROR)
logging.getLogger("asyncio").setLevel(logging.CRITICAL)



import json
from typing import Any

from .factool_utils.chat_api import OpenAIChat
from .factool_utils.search_api import GoogleSerperAPIWrapper
from .factool_utils.prompt import QUERY_GENERATION_PROMPT

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("factool_evidence_retriever", "claims", "evidences")
class FactoolEvidenceRetriever(StandardTaskSolver):
    """
    A solver to retrieve evidences for a list of evidence. (online content + its sources) for a list of claims.
    """
    def __init__(self, args):
        super().__init__(args)
        self.gpt_model = self.global_config.get("llm_in_use", "gpt-4o")
        self.gpt = OpenAIChat(self.gpt_model)
        
        self.path_save_evidence = args.get("path_save_evidence", "evidence.json")
        self.queries = None
        self.search_outputs_for_claims = None

        self.query_prompt = QUERY_GENERATION_PROMPT

        self.search_engine = GoogleSerperAPIWrapper(snippet_cnt=10)
        

    # async def coro_queries (self, factool_instance, claims_in_response):
    #    self.queries = await factool_instance.pipelines["kbqa_online"]._query_generation(claims_in_response)
    # async def coro_search_outputs_for_claims (self, factool_instance):
    #    self.search_outputs_for_claims = await factool_instance.pipelines["kbqa_online"].tool.run(self.queries)
    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        queries = self._query_generation(claims=claims)
        search_outputs_for_claims = self.search_engine.run(queries)
        
        
        evidences: dict[str, dict[str, Any]] = {}
        for i, claim in enumerate(claims):
            evidence_list: list[dict] = []
            for j, search_outputs_for_claim in enumerate(
                search_outputs_for_claims[i]
            ):
                evidence_list.append(
                    {
                        "evidence_id": j,
                        "web_page_snippet_manual": search_outputs_for_claim["content"],
                        "query": [queries[i]],
                        "url": search_outputs_for_claim["source"],
                        "web_text": [],
                    }
                )
            evidences[claim] = {
                "claim": claim,
                "automatic_queries": queries[i],
                "evidence_list": evidence_list,
            }

        # write to json file
        # Serializing json
        json_object = json.dumps(evidences, indent=4)

        # Writing to sample.json
        with open(self.path_save_evidence, "w") as outfile:
            outfile.write(json_object)

        state.set(self.output_name, evidences)
        return True, state

    def _query_generation(self, claims):
        messages_list = [
            [
                {"role": "system", "content": self.query_prompt["system"]},
                {
                    "role": "user",
                    "content": self.query_prompt["user"].format(input=claim),
                },
            ]
            for claim in claims
        ]
        return self.gpt.run(messages_list, list)



from .rarr_utils.hallucination import run_evidence_hallucination
from .prompts.hallucination_prompts import EVIDENCE_HALLUCINATION

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver


@Solver.register("llm_retriever", "claims_with_questions", "claims_with_evidences")
class RARRLLMRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("model", "gpt-4o-instruct")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        for claim, contents in claims.items():
            questions = contents.get("questions", [])
            evidences = []
            for question in questions:
                evidences.append(run_evidence_hallucination(question, model=self.model, prompt=EVIDENCE_HALLUCINATION))
            claims[claim]["evidences"] = evidences

        state.set(self.output_name, claims)
        return True, state



from typing import Any, Optional, TypeVar, Type

from botocore.client import BaseClient
from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
from botocore.exceptions import ClientError

from openfactcheck.utils import logging
from openfactcheck.services.model import DynamoDBBaseModel

logger = logging.get_logger(__name__)

T = TypeVar("T", bound=DynamoDBBaseModel)


class DynamoDBInterface:
    """
    Interface to interact with AWS DynamoDB service.

    Provides methods to create, update, and fetch items from a DynamoDB table.

    Parameters
    ----------
    dynamodb_svc : botocore.client.BaseClient
        Boto3 DynamoDB service client.
    table : str
        Name of the DynamoDB table.

    Attributes
    ----------
    logger : logging.Logger
        Logger instance for the class.
    table : str
        Name of the DynamoDB table.
    dynamodb_svc : botocore.client.BaseClient
        Boto3 DynamoDB service client.

    Methods
    -------
    create_or_update(storable: DynamoDBBaseModel) -> None:
        Create or update a DynamoDB item based on the provided storable object.
    fetch(pk: str, model: Type[T]) -> Optional[T]:
        Fetch a DynamoDB item by primary key and deserialize it into the provided model.

    Raises
    ------
    ClientError
        If the DynamoDB service reports an error.
    """

    def __init__(self, dynamodb_svc: BaseClient, table: str) -> None:
        self.logger = logger
        self.table: str = table
        self.dynamodb_svc: BaseClient = dynamodb_svc

    def _serialize_item(self, storable: DynamoDBBaseModel) -> dict[str, Any]:
        """
        Serialize a DynamoDBBaseModel instance to a dictionary format for DynamoDB storage.

        Parameters
        ----------
        storable : DynamoDBBaseModel
            The object to serialize.

        Returns
        -------
        Dict[str, Any]
            The serialized item ready to be stored in DynamoDB.
        """
        serializer = TypeSerializer()
        item_dict = storable.model_dump(exclude_unset=True, by_alias=True)
        av = {k: serializer.serialize(v) for k, v in item_dict.items()}

        # Add the primary key
        av["PK"] = serializer.serialize(storable.PK)

        # Optionally add the sort key, if present
        if storable.SK is not None:
            av["SK"] = serializer.serialize(storable.SK)

        # Optionally add the GS1 partition key, if present
        if storable.GS1PK is not None:
            av["GS1PK"] = serializer.serialize(storable.GS1PK)

        return av

    def _deserialize_item(self, item: dict[str, Any], model: Type[T]) -> T:
        """
        Deserialize a DynamoDB item into an instance of the provided model.

        Parameters
        ----------
        item : dict
            The DynamoDB item to deserialize.
        model : Type[T]
            The model class to instantiate with the deserialized data.

        Returns
        -------
        T
            An instance of the model class populated with data from the item.
        """
        deserializer = TypeDeserializer()
        attributes = {k: deserializer.deserialize(v) for k, v in item.items()}
        return model(**attributes)

    def _paged_scan(self) -> list[dict[str, Any]]:
        """
        Perform a paginated scan of the DynamoDB table and return all items.

        Returns
        -------
        list of dict
            A list of items retrieved from the DynamoDB table.

        Raises
        ------
        ClientError
            If the DynamoDB service reports an error.
        """
        try:
            items = []
            scan_kwargs = {"TableName": self.table}
            while True:
                response = self.dynamodb_svc.scan(**scan_kwargs)
                items.extend(response.get("Items", []))
                self.logger.debug(f"Fetched {len(response.get('Items', []))} items in this page.")
                if "LastEvaluatedKey" in response:
                    scan_kwargs["ExclusiveStartKey"] = response["LastEvaluatedKey"]
                else:
                    break
            self.logger.info(f"Total {len(items)} items fetched from table {self.table}.")
            return items
        except ClientError as e:
            self.logger.error(f"DynamoDBInterface._paged_scan failed: {e}")
            raise

    def create_or_update(self, storable: DynamoDBBaseModel) -> None:
        """
        Create or update a DynamoDB item based on the provided storable object.

        Parameters
        ----------
        storable : DynamoDBBaseModel
            The object to create or update in DynamoDB.

        Raises
        ------
        ClientError
            If the DynamoDB service reports an error.
        """
        try:
            item = self._serialize_item(storable)
            self.dynamodb_svc.put_item(TableName=self.table, Item=item)
            self.logger.info(f"Item with PK={storable.PK} created/updated successfully.")
        except ClientError as e:
            self.logger.error(f"DynamoDBInterface.create_or_update failed: {e}")
            raise

    def fetch(self, pk: str, model: Type[T]) -> Optional[T]:
        """
        Fetch a DynamoDB item by primary key and deserialize it into the provided model.

        Parameters
        ----------
        pk : str
            The primary key of the item to fetch.
        model : Type[T]
            The model class to deserialize the item into.

        Returns
        -------
        Optional[T]
            An instance of the model if found; otherwise, None.

        Raises
        ------
        ClientError
            If the DynamoDB service reports an error.
        """
        try:
            key = {"PK": {"S": pk}}
            response = self.dynamodb_svc.get_item(TableName=self.table, Key=key)
            if "Item" not in response:
                self.logger.info(f"No item found with PK={pk}.")
                return None
            self.logger.info(f"Item with PK={pk} fetched successfully.")
            return self._deserialize_item(response["Item"], model)
        except ClientError as e:
            self.logger.error(f"DynamoDBInterface.fetch failed: {e}")
            raise

    def delete(self, pk: str) -> None:
        """
        Delete a DynamoDB item by primary key.

        Parameters
        ----------
        pk : str
            The primary key of the item to delete.

        Raises
        ------
        ClientError
            If the DynamoDB service reports an error.
        """
        try:
            key = {"PK": {"S": pk}}
            self.dynamodb_svc.delete_item(TableName=self.table, Key=key)
            self.logger.info(f"Item with PK={pk} deleted successfully.")
        except ClientError as e:
            self.logger.error(f"DynamoDBInterface.delete failed: {e}")
            raise

    def list(self, model: Type[T]) -> Optional[list[T]]:
        """
        List all items in the DynamoDB table and deserialize them into the provided model.

        Parameters
        ----------
        model : Type[T]
            The model class to deserialize the items into.

        Returns
        -------
        Optional[List[T]]
            A list of instances of the model class if items are found; otherwise, None.

        Raises
        ------
        ClientError
            If the DynamoDB service reports an error.
        """
        try:
            items = self._paged_scan()
            if not items:
                self.logger.info(f"No items found in table {self.table}.")
                return None
            self.logger.info("Items fetched successfully.")
            return [self._deserialize_item(item, model) for item in items]
        except ClientError as e:
            self.logger.error(f"DynamoDBInterface.list failed: {e}")
            raise



from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("rarr_concat_response_generator", "revised_claims", "output")
class RARRConcatResponseRegenerator(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        revised_document = " ".join(list(claims.values())).strip()
        # print(revised_document)
        state.set(self.output_name, revised_document)
        return True, state



from __future__ import annotations

import os
import json
import ast
import openai
import asyncio
from openai import AsyncOpenAI
from anthropic import AsyncAnthropic, HUMAN_PROMPT, AI_PROMPT


class OpenAIChat:
    def __init__(
        self,
        model_name,
        max_tokens=2500,
        temperature=0,
        top_p=1,
        request_timeout=20,
    ):
        if "gpt" not in model_name:
            openai.api_base = "http://localhost:8000/v1"
        else:
            # openai.api_base = "https://api.openai.com/v1"
            openai.api_key = os.environ.get("OPENAI_API_KEY", None)
            assert (
                openai.api_key is not None
            ), "Please set the OPENAI_API_KEY environment variable."
            assert (
                openai.api_key != ""
            ), "Please set the OPENAI_API_KEY environment variable."
        self.client = AsyncOpenAI()

        self.config = {
            "model_name": model_name,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "request_timeout": request_timeout,
        }

    def extract_list_from_string(self, input_string):
        start_index = input_string.find("[")
        end_index = input_string.rfind("]")

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index : end_index + 1]
        else:
            return None

    def extract_dict_from_string(self, input_string):
        start_index = input_string.find("{")
        end_index = input_string.rfind("}")

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index : end_index + 1]
        else:
            return None

    def _json_fix(self, output):
        return output.replace("```json\n", "").replace("\n```", "")

    def _boolean_fix(self, output):
        return output.replace("true", "True").replace("false", "False")

    def _type_check(self, output, expected_type):
        try:
            output_eval = ast.literal_eval(output)
            if not isinstance(output_eval, expected_type):
                print(
                    f"Type mismatch: expected {expected_type}, got {type(output_eval)}"
                )
                return None
            return output_eval
        except:
            if expected_type == str:
                return output
            else:
                print(f"Error evaluating output: {output}")
                return None

    async def dispatch_openai_requests(
        self,
        messages_list,
    ) -> list[str]:
        """Dispatches requests to OpenAI API asynchronously.

        Args:
            messages_list: List of messages to be sent to OpenAI ChatCompletion API.
        Returns:
            List of responses from OpenAI API.
        """

        async def _request_with_retry(messages, retry=3):
            for _ in range(retry):
                try:
                    response = await asyncio.wait_for(
                        self.client.chat.completions.create(
                            model=self.config["model_name"],
                            messages=messages,
                            max_tokens=self.config["max_tokens"],
                            temperature=self.config["temperature"],
                            top_p=self.config["top_p"],
                        ),
                        timeout=self.config["request_timeout"],  # seconds
                    )
                    return response
                except asyncio.TimeoutError:
                    print("Timeout reached for request. Retrying...")
                    await asyncio.sleep(1)
                except openai.RateLimitError:
                    await asyncio.sleep(1)
                except openai.Timeout:
                    await asyncio.sleep(1)
                except openai.APIError:
                    await asyncio.sleep(1)
            return None

        async_responses = [_request_with_retry(messages) for messages in messages_list]

        return await asyncio.gather(*async_responses, return_exceptions=True)

    def run(self, messages_list, expected_type):
        retry = 1
        responses = [None for _ in range(len(messages_list))]
        messages_list_cur_index = [i for i in range(len(messages_list))]

        while retry > 0 and len(messages_list_cur_index) > 0:
            messages_list_cur = [messages_list[i] for i in messages_list_cur_index]

            predictions = asyncio.run(
                self.dispatch_openai_requests(
                    messages_list=messages_list_cur,
                )
            )

            # Save the cost of the API call to a JSONL file
            if os.environ.get("SAVE_MODEL_COST", "False") == "True":
                MODEL_COST_PATH = os.environ.get("MODEL_COST_PATH", "model_cost.jsonl")
                for prediction in predictions:
                    if prediction is not None:
                        if hasattr(prediction, "usage"):
                            completion_tokens = prediction.usage.completion_tokens
                            prompt_tokens = prediction.usage.prompt_tokens
                            total_tokens = prediction.usage.total_tokens
                            with open(MODEL_COST_PATH, "a") as f:
                                f.write(
                                    json.dumps(
                                        {
                                            "model": self.config["model_name"],
                                            "prompt_tokens": prompt_tokens,
                                            "completion_tokens": completion_tokens,
                                            "total_tokens": total_tokens,
                                        }
                                    )
                                    + "\n"
                                )

            preds = [
                self._type_check(
                    self._boolean_fix(
                        self._json_fix(prediction.choices[0].message.content)
                    ),
                    expected_type,
                )
                if prediction is not None and hasattr(prediction, "choices")
                else None
                for prediction in predictions
            ]
            finised_index = []
            for i, pred in enumerate(preds):
                if pred is not None:
                    responses[messages_list_cur_index[i]] = pred
                    finised_index.append(messages_list_cur_index[i])

            messages_list_cur_index = [
                i for i in messages_list_cur_index if i not in finised_index
            ]

            retry -= 1

        return responses


class AnthropicChat:
    def __init__(
        self,
        model_name: str,
        max_tokens: int = 2500,
        temperature: float = 0,
        top_p: float = 1,
        request_timeout: float = 20,
    ):
        api_key = os.environ.get("ANTHROPIC_API_KEY")
        assert api_key, "Please set the ANTHROPIC_API_KEY environment variable."

        self.client = AsyncAnthropic(api_key=api_key)
        self.config = {
            "model_name": model_name,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "request_timeout": request_timeout,
        }

    def extract_list_from_string(self, input_string: str) -> str | None:
        start = input_string.find("[")
        end = input_string.rfind("]")
        if start != -1 and end != -1 and start < end:
            return input_string[start : end + 1]
        return None

    def extract_dict_from_string(self, input_string: str) -> str | None:
        start = input_string.find("{")
        end = input_string.rfind("}")
        if start != -1 and end != -1 and start < end:
            return input_string[start : end + 1]
        return None

    def _json_fix(self, output: str) -> str:
        if isinstance(output, str):
            return output.replace("```json\n", "").replace("\n```", "")
        else:
            return output

    def _boolean_fix(self, output: str) -> str:
        if isinstance(output, str):
            return output.replace("true", "True").replace("false", "False")
        else:
            return output

    def _type_check(self, output: str, expected_type: type):
        try:
            val = ast.literal_eval(output)
            if not isinstance(val, expected_type):
                print(f"Type mismatch: expected {expected_type}, got {type(val)}")
                return None
            return val
        except Exception:
            if expected_type == str:
                return output
            print(f"Error evaluating output: {output}")
            return None

    async def dispatch_anthropic_requests(
        self,
        messages_list: list[list[dict]],
    ) -> list[object | None]:
        """Send batches via the Messages API with retries."""

        async def _request_with_retry(
            messages: list[dict], retry: int = 3
        ) -> object | None:
            # Extract any system prompt to top‐level
            system_content = None
            filtered = []
            for msg in messages:
                if msg.get("role") == "system":
                    system_content = msg["content"]
                else:
                    filtered.append({"role": msg["role"], "content": msg["content"]})

            for _ in range(retry):
                try:
                    return await asyncio.wait_for(
                        self.client.messages.create(
                            model=self.config["model_name"],
                            system=system_content,
                            messages=filtered,
                            max_tokens=self.config["max_tokens"],
                            temperature=self.config["temperature"],
                            top_p=self.config["top_p"],
                        ),
                        timeout=self.config["request_timeout"],
                    )
                except asyncio.TimeoutError:
                    print("Anthropic request timed out, retrying…")
                    await asyncio.sleep(1)
                except Exception as e:
                    print(f"Anthropic API error ({e}), retrying…")
                    await asyncio.sleep(1)
            return None

        tasks = [_request_with_retry(msgs) for msgs in messages_list]
        return await asyncio.gather(*tasks, return_exceptions=False)

    def run(
        self,
        messages_list: list[list[dict]],
        expected_type: type,
    ) -> list[object | None]:
        """Dispatch messages and type‐check their responses."""
        responses = [None] * len(messages_list)
        pending_idx = list(range(len(messages_list)))
        attempts = 1

        while attempts > 0 and pending_idx:
            batch = [messages_list[i] for i in pending_idx]
            completions = asyncio.run(self.dispatch_anthropic_requests(batch))
            finished = []

            for idx_in_batch, comp in enumerate(completions):
                if comp is None or not hasattr(comp, "content"):
                    continue

                raw = comp.content
                # Optional cost logging
                if os.environ.get("SAVE_MODEL_COST", "False") == "True" and hasattr(
                    comp, "usage"
                ):
                    MODEL_COST_PATH = os.environ.get(
                        "MODEL_COST_PATH", "model_cost.jsonl"
                    )
                    with open(MODEL_COST_PATH, "a") as f:
                        f.write(
                            json.dumps(
                                {
                                    "model": self.config["model_name"],
                                    "input_tokens": comp.usage.input_tokens,
                                    "output_tokens": comp.usage.output_tokens,
                                    "total_tokens": comp.usage.input_tokens
                                    + comp.usage.output_tokens,
                                }
                            )
                            + "\n"
                        )

                # Parse TextBox list
                raw_text = ""
                for i in range(len(raw)):
                    raw_text += raw[i].text

                cleaned = self._boolean_fix(self._json_fix(raw_text))
                result = self._type_check(cleaned, expected_type)
                if result is not None:
                    real_idx = pending_idx[idx_in_batch]
                    responses[real_idx] = result
                    finished.append(real_idx)

            pending_idx = [i for i in pending_idx if i not in finished]
            attempts -= 1

        return responses



from pydantic import BaseModel
from abc import ABC, abstractmethod
from typing import Optional


class DynamoDBBaseModel(BaseModel, ABC):
    """
    Base model for DynamoDB items, requiring a primary key (PK).
    Optional sort key (SK) and global secondary index partition key (GS1PK) can be defined.
    """

    @property
    @abstractmethod
    def PK(self) -> str:
        """Primary Key for DynamoDB storage. Must be implemented by subclasses."""
        pass

    @property
    def SK(self) -> Optional[str]:
        """Sort Key for DynamoDB storage. Optional."""
        return None

    @property
    def GS1PK(self) -> Optional[str]:
        """Global Secondary Index 1 Partition Key for DynamoDB. Optional."""
        return None

    class Config:
        # Updated configuration key for Pydantic v2
        populate_by_name = True

        # Allow extra fields (useful for DynamoDB metadata)
        extra = "allow"



from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("concat_response_generator", "claim_info", "output")
class ConcatResponseRegenerator(StandardTaskSolver):
    """
    A solver to concatenate the edited claims into a single document.
    """
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        edited_claims = [v["edited_claims"] for _, v in claim_info.items()]
        revised_document = " ".join(edited_claims).strip()
        state.set(self.output_name, revised_document)
        return True, state



import os
import json
import asyncio
import aiohttp

from .chat_api import OpenAIChat, AnthropicChat
from .prompt import (
    URDU_TO_ENGLISH_TRANSLATION_PROMPT,
    ENGLISH_TO_URDU_TRANSLATION_PROMPT,
)


class GoogleSerperAPIWrapper:
    """Wrapper around the Serper.dev Google Search API.
    You can create a free API key at https://serper.dev.
    To use, you should have the environment variable ``SERPER_API_KEY``
    set with your API key, or pass `serper_api_key` as a named parameter
    to the constructor.
    Example:
        .. code-block:: python
            from langchain import GoogleSerperAPIWrapper
            google_serper = GoogleSerperAPIWrapper()
    """

    def __init__(self, snippet_cnt=10) -> None:
        self.k = snippet_cnt
        self.gl = "us"
        self.hl = "ur"
        self.serper_api_key = os.environ.get("SERPER_API_KEY", None)
        assert (
            self.serper_api_key is not None
        ), "Please set the SERPER_API_KEY environment variable."
        assert (
            self.serper_api_key != ""
        ), "Please set the SERPER_API_KEY environment variable."

        self.gpt_model = os.environ.get("MODEL_NAME", "gpt-4o")
        if "claude" in self.gpt_model:
            self.gpt = AnthropicChat(self.gpt_model)
        else:
            self.gpt = OpenAIChat(self.gpt_model)
        self.english_to_urdu_translation_prompt = ENGLISH_TO_URDU_TRANSLATION_PROMPT
        self.urdu_to_english_translation_prompt = URDU_TO_ENGLISH_TRANSLATION_PROMPT

    async def _google_serper_search_results(
        self, session, search_term: str, gl: str, hl: str
    ) -> dict:
        headers = {
            "X-API-KEY": self.serper_api_key or "",
            "Content-Type": "application/json",
        }
        payload = {"q": search_term, "gl": gl, "hl": hl}
        try:
            async with session.post(
                "https://google.serper.dev/search",
                headers=headers,
                json=payload,  # <-- this is correct for POST JSON
                timeout=aiohttp.ClientTimeout(total=15),
            ) as response:
                response.raise_for_status()
                return await response.json()
        except Exception as e:
            print(f"Error for query '{search_term}': {e}")
            return {}

    def _parse_results(self, results):
        snippets = []

        if os.environ.get("SAVE_SERPER_COST", "False") == "True":
            SERPER_COST_PATH = os.environ.get("SERPER_COST_PATH", "serper_cost.jsonl")
            if results.get("credits"):
                credits = results.get("credits")
                with open(SERPER_COST_PATH, "a") as f:
                    f.write(json.dumps({"google_serper_credits": credits}) + "\n")

        if results.get("answerBox"):
            answer_box = results.get("answerBox", {})
            if answer_box.get("answer"):
                element = {"content": answer_box.get("answer"), "source": "None"}
                return [element]
            elif answer_box.get("snippet"):
                element = {
                    "content": answer_box.get("snippet").replace("\n", " "),
                    "source": "None",
                }
                return [element]
            elif answer_box.get("snippetHighlighted"):
                element = {
                    "content": answer_box.get("snippetHighlighted"),
                    "source": "None",
                }
                return [element]

        if results.get("knowledgeGraph"):
            kg = results.get("knowledgeGraph", {})
            title = kg.get("title")
            entity_type = kg.get("type")
            if entity_type:
                element = {"content": f"{title}: {entity_type}", "source": "None"}
                snippets.append(element)
            description = kg.get("description")
            if description:
                element = {"content": description, "source": "None"}
                snippets.append(element)
            for attribute, value in kg.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": "None"}
                snippets.append(element)

        for result in results["organic"][: self.k]:
            if "snippet" in result:
                element = {"content": result["snippet"], "source": result["link"]}
                snippets.append(element)
            for attribute, value in result.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": result["link"]}
                snippets.append(element)

        if len(snippets) == 0:
            element = {
                "content": "No good Google Search Result was found",
                "source": "None",
            }
            return [element]

        # keep only the first k snippets
        snippets = snippets[: int(self.k / 2)]

        return snippets

    async def parallel_searches(self, search_queries, gl, hl):
        async with aiohttp.ClientSession() as session:
            tasks = [
                self._google_serper_search_results(session, query, gl, hl)
                for query in search_queries
            ]
            search_results = await asyncio.gather(*tasks, return_exceptions=True)
            return search_results

    def run(self, queries):
        """Run query through GoogleSearch and parse result."""
        flattened_queries = []

        for sublist in queries:
            if sublist is None:
                sublist = ["None", "None"]
            for item in sublist:
                flattened_queries.append(item)

        # Get results
        results = asyncio.run(
            self.parallel_searches(flattened_queries, gl=self.gl, hl=self.hl)
        )
        snippets_list = []
        for i in range(len(results)):
            snippets_list.append(self._parse_results(results[i]))

        # Flatten the list of snippets
        snippets_split = [
            snippets_list[i] + snippets_list[i + 1]
            for i in range(0, len(snippets_list), 2)
        ]

        snippets_split_length = 0
        for snippet_split in snippets_split:
            if snippets_split_length == 0:
                snippets_split_length = len(snippet_split)
            if snippets_split_length > len(snippet_split):
                snippets_split_length = len(snippet_split)

        # Check if the evidence threshold is met
        print(f"Evidence threshold is set to {os.environ.get('EVIDENCE_THRESHOLD', 5)}")
        if snippets_split_length <= int(os.environ.get("EVIDENCE_THRESHOLD", 5)):
            print(f"Evidence threshold not met: {snippets_split_length}")
            # Translate Queries to English
            messages_list = [
                [
                    {
                        "role": "system",
                        "content": self.urdu_to_english_translation_prompt["system"],
                    },
                    {
                        "role": "user",
                        "content": self.urdu_to_english_translation_prompt[
                            "user"
                        ].format(input=query),
                    },
                ]
                for query in flattened_queries
            ]
            english_queries = self.gpt.run(messages_list, str)

            # Get results in English Language
            results = asyncio.run(
                self.parallel_searches(english_queries, gl=self.gl, hl="en")
            )
            snippets_list_en = []
            for i in range(len(results)):
                snippets_list_en.append(self._parse_results(results[i]))

            # Flatten the list of snippets
            snippets_split_en = [
                snippets_list_en[i] + snippets_list_en[i + 1]
                for i in range(0, len(snippets_list_en), 2)
            ]

            translated_snippets = []
            for snippet_split in snippets_split_en:
                messages_list = [
                    [
                        {
                            "role": "system",
                            "content": self.english_to_urdu_translation_prompt[
                                "system"
                            ],
                        },
                        {
                            "role": "user",
                            "content": self.english_to_urdu_translation_prompt[
                                "user"
                            ].format(input=snippet["content"]),
                        },
                    ]
                    for snippet in snippet_split
                ]
                urdu_snippets = self.gpt.run(messages_list, str)
                translated_snippet = []
                for urdu_snippet in urdu_snippets:
                    translated_snippet.append({"content": urdu_snippet})
                translated_snippets.append(translated_snippet)

            # Combine the translated snippets with the original snippets
            combined_snippets = [
                list1 + list2
                for list1, list2 in zip(snippets_split, translated_snippets)
            ]
            return combined_snippets
        else:
            return snippets_split



import nltk
import spacy

from openfactcheck.state import FactCheckerState
from openfactcheck.solver import StandardTaskSolver, Solver

from .factcheckgpt_utils.openai_api import gpt
from .factcheckgpt_utils.data_util import save_to_file
from .factcheckgpt_utils.prompt import (
    DOC_TO_INDEPEDENT_SENTENCES_PROMPT,
    SENTENCES_TO_CLAIMS_PROMPT,
    DOC_TO_SENTENCES_PROMPT,
    CHECKWORTHY_PROMPT_BOOL,
    SPECIFY_CHECKWORTHY_CATEGORY_PROMPT,
)


@Solver.register("factcheckgpt_claimprocessor", "response", "claims")
class FactCheckGPTClaimProcessor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("factcheckgpt_model", "gpt-4o")
        self.num_retries = self.global_config.get("num_retries", 3)
        self.mode = args.get("mode", "independent_sentences")
        self.decompose_system_role = "You are good at decomposing and decontextualizing text."
        self.worthines_filter_system_role = "You are a helpful factchecker assistant."
        self.rule_based_method = args.get("rule_based_tool", "spacy")
        self.spacy_model = args.get("spacy_model", "en_core_web_sm")
        self.prompt = {
            "sentences": DOC_TO_SENTENCES_PROMPT,
            "independent_sentences": DOC_TO_INDEPEDENT_SENTENCES_PROMPT,
            "claims": SENTENCES_TO_CLAIMS_PROMPT,
        }.get(self.mode, DOC_TO_INDEPEDENT_SENTENCES_PROMPT)
        nlp = spacy.load(self.spacy_model)
        self.rule_based_tool = {
            "nltk": lambda x: [x.strip() for x in nltk.sent_tokenize(x) if len(x.strip()) >= 3],
            "spacy": lambda x: [x.text.strip() for x in nlp(x).sents if len(x.text.strip()) >= 3],
        }.get(self.rule_based_method, "nltk")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        # We have merged the text decomposer and worthiness filter here.
        response = state.get(self.input_name)
        claims = [response]

        user_input = self.prompt.format(doc=response).strip()
        r = gpt(user_input, model=self.model, system_role=self.decompose_system_role, num_retries=self.num_retries)
        try:
            claims = eval(r)
        except Exception as e:
            print(f"An unexpected error occurred: {e}.")
            save_to_file(r)

        if not isinstance(claims, list):
            print(
                f"{self.model} output {r}. It does not output a list of sentences correctly, return rule-based split results."
            )
            claims = self.rule_based_tool(response)

        worthiness = [True] * len(claims)
        user_input = CHECKWORTHY_PROMPT_BOOL.format(claims=claims)
        response = gpt(
            user_input, model=self.model, system_role=self.worthines_filter_system_role, num_retries=self.num_retries
        )
        # TODO refine check worthiness prompt, value returned not reasonable.
        try:
            worthiness = eval(response)
            assert len(worthiness) == len(claims)
        except AssertionError as e:
            print(f"An unexpected error occurred: {e}")
            print(f"There are {len(claims)} texts, while {len(worthiness)} checkworthy predictions.")
            return False, state
        except Exception as e:
            print(f"An unexpected error occurred: {e}")
            return False, state

        valid_claims = list(map(lambda x: x[1], filter(lambda x: x[0], zip(worthiness, claims))))
        state.set(self.output_name, valid_claims)
        return True, state
